import multipletau
import matplotlib.pyplot as plt
import numpy as np
import re
import os
import ntpath
import scipy as spy
from pathlib import Path
import fnmatch

from .extract_spad_data import extract_spad_data
from .distance2detelements import distance2detelements, spad_shift_vector_crosscorr
from .distance2detelements import spad_coord_from_det_numb as coord
from .get_fcs_info import get_file_info, get_metafile_from_file
from .meas_to_count import file_to_fcs_count, czi2h5
from .atimes2corr import atimes_2_corr
from .timetrace2corr import tt2corr
from .corr2csv import corr2csv

from ..tools.plot_colors import plot_colors
from ..tools.list_files import list_files
from ..tools.bindata import bindata_chunks
from ..tools.calcdist_from_coord import list_of_pixel_pairs_at_distance


class Correlations:
    pass

def fcs_sparse_matrices(fname, accuracy=16, split=10, time_trace=False, return_obj=False, averaging=None, root=0):
    """
    Calculate correlations using spare matrices properties.
    Author: Eleonora Perego

    Parameters
    ----------
    fname : string
        file name with fcs data.
    accuracy : int, optional
        Accuracy of the autocorrelation function. The default is 16.
    split : float, optional
        Number of seconds of each chunk to split the data into
        E.g. split=10 will divide a 60 second stream in 6 ten-second
        traces and calculate G for each  ividual trace. The default is 10.
    time_trace : boolean, optional
        see output. The default is False.
    return_obj : boolean, optional
        Return data as 4D np.array() or object. The default is False.
    root : int, optional
        Used for GUI only. The default is 0.
    averaging : list of list of two strings
        Used for averaging multiple cross-correlations (for each chunk)
        E.g. for having G.up_chunk0 = mean(det12x7_chunk0 and det11x6_chunk0)
            and G.down_chunk0 = mean(det12x17_chunk0 and det11x16_chunk0)
            etc.
        Use: averaging = [['up', '12x7+11x6'], ['down', '12x17+11x16']]

    Returns
    -------
    G : object
        np.array or object with autocorrelations.
    data : np.array()
        if time_trace == False: last chunk of raw data
        if time_trace == True: full time trace binned: [Nx25] array.

    """
    
    
    metafile = get_metafile_from_file(fname)
    info = get_file_info(metafile)
    
    dwellTime = 1e-6 * info.timeResolution
    duration = info.duration
    
    N = int(np.floor(duration / split)) # number of chunks

    G = Correlations()
    G.dwellTime = dwellTime
    chunkSize = int(np.floor( split/ dwellTime))
    Gcorrs = []
    
    for chunk in range(N):
        # --------------------- CALCULATE CORRELATIONS SINGLE CHUNK ---------------------
        print("+-----------------------")
        print("| Loading chunk " + str(chunk))
        print("+-----------------------")
        
        if root != 0:
            root.progress = chunk / N
        
        data = file_to_fcs_count(fname, np.uint8, chunkSize, chunk*chunkSize)
        
        if time_trace == True:
            binSize = int(chunkSize / 1000 * N) # time trace binned to 1000 data points
            bdata = bindata_chunks(data, binSize)
            bdataL = len(bdata)
            num_ch = np.shape(bdata)[1]
            if chunk == 0:
                timetraceAll = np.zeros((1000, num_ch), dtype=int)
            timetraceAll[chunk*bdataL:(chunk+1)*bdataL, :] = bdata 
        
        Gsplit,Gtimes = fcs_sparse(data, dwellTime, accuracy)
        Gcorrs.append(Gsplit)
    G.lagtimes = Gtimes
    G.Gsplit   = Gcorrs
    
    if time_trace:
        data = timetraceAll
    
    if return_obj:
        # convert 4D numpy array (chunk, y, corr1, corr2) to corr object
        Garr = np.asarray(G.Gsplit)
        Gshape = np.shape(Garr)
        Nchunk = Gshape[0]
        Ntau = Gshape[1]
        Nx = Gshape[2]
        Ny = Gshape[3]
        Gout = Correlations()
        if averaging is None:
            for c in range(Nchunk):
                for x in range(Nx):
                    for y in range(Ny):
                        Gtemp = np.zeros((Ntau, 2))
                        Gtemp[:,0] = np.squeeze(G.lagtimes)
                        Gtemp[:,1] = Garr[c, :, x, y]
                        setattr(Gout, "det" + str(x) + "x" + str(y) + "_chunk" + str(c), Gtemp)
        else:
            # average over multiple cross-correlations
            avs = [i[1] for i in averaging] # which correlations to average
            els = [i[0] for i in averaging] # their new names
            for c in range(Nchunk):
                for el, av in enumerate(avs):
                    singleAv = [int(i) for i in re.findall(r'\d+', av)]
                    Nav = int(len(singleAv) / 2)
                    Gtemp = np.zeros((len(G.lagtimes), 2))
                    for i in range(Nav):
                        Gtemp[:,0] += np.squeeze(G.lagtimes)
                        Gtemp[:,1] += Garr[c, :, singleAv[2*i], singleAv[2*i+1]]
                    Gtemp /= Nav
                    setattr(Gout, els[el] + "_chunk" + str(c), Gtemp)
                
        Gout = fcs_av_chunks(Gout, list(range(Nchunk)))
        Gout.dwellTime = dwellTime
        G = Gout
    
    return G, data

def fcs_sparse(data, dwell_time, m=50, normalize = True):
    # object from correlations class in which all correlation data is stored
    G = Correlations()
    
    # dwell time
    G.dwellTime = dwell_time
    
    # Check parameters
    if m // 2 != m / 2:
        mold = m
        m = np.int_((m // 2 + 1) * 2)
    else:
        m = np.int_(m)

    N = N0 = data.shape[0]
    Nchan = data.shape[1]
    k = np.int_(np.floor(np.log2(N / m)))
    lenG = m + k * m // 2 + 1
    Gtimes = np.zeros((lenG, 1), dtype = "float32")
    G = np.zeros((lenG, Nchan, Nchan), dtype="float32")
    normstat = np.zeros(lenG, dtype="float32")
    normnump = np.zeros(lenG, dtype="float32")
    
    spdata = spy.sparse.csr_matrix(data.transpose())
    spdata = spdata.astype("float32")
    
    traceavg = spdata.mean(axis = 1)
    traceavg[traceavg==0]=1
    
    # Calculate autocorrelation function for first m+1 bins
    for n in range(0, m + 1):
        Gtimes[n] = dwell_time * n
        res = spdata[:,:N-n].dot(spdata[:,n:].transpose())
        G[n] = res.toarray()
        normstat[n] = N - n
        normnump[n] = N
    
    if N % 2 == 1:
        N -= 1
    # compress every second element
    spdata = (spdata[:,:N:2] + spdata[:,1:N:2]) / 2
    spdata = spdata.toarray() # spdata is now full
    
    spdata = spdata-traceavg
    
    N //= 2
    
    # Start iteration for each m/2 values
    for step in range(1, k + 1):
        # Get the next m/2 values via correlation of the trace
        for n in range(1, m // 2 + 1):
            npmd2 = n + m // 2
            idx = m + n + (step - 1) * m // 2
            if spdata[:,:N - npmd2].shape[1] == 0:
                # This is a shortcut that stops the iteration once the
                # length of the trace is too small to compute a corre-
                # lation. The actual length of the correlation function
                # does not only depend on k - We also must be able to
                # perform the sum with respect to k for all elements.
                # For small N, the sum over zero elements would be
                # computed here.
                #
                # One could make this for-loop go up to maxval, where
                #   maxval1 = int(m/2)
                #   maxval2 = int(N-m/2-1)
                #   maxval = min(maxval1, maxval2)
                # However, we then would also need to find out which
                # element in G is the last element...
                G = G[:idx - 1]
                normstat = normstat[:idx - 1]
                normnump = normnump[:idx - 1]
                # Note that this break only breaks out of the current
                # for loop. However, we are already in the last loop
                # of the step-for-loop. That is because we calculated
                # k in advance.
                break
            else:
                Gtimes[idx] = dwell_time * npmd2 * 2**step
                # This is the computationally intensive step
                G[idx] = spdata[:,:N-npmd2].dot(spdata[:,npmd2:].transpose())
                normstat[idx] = N - npmd2
                normnump[idx] = N
        # Check if len(trace) is even:
        if N % 2 == 1:
            N -= 1
        # compress every second element
        spdata = (spdata[:,:N:2] + spdata[:,1:N:2]) / 2  
        N //= 2
   
    if normalize:
        # G /= normstat.reshape(lenG,1,1)
        # G /= (traceavg*traceavg.transpose())
        # G -= 1
        lenG = len(G)
        G /= normstat.reshape(lenG,1,1)
        G /= (traceavg*traceavg.transpose())
        G[1:m+1] -= 1

    return G, Gtimes
    

def correlate_parallel(a, m=16, deltat=1, normalize=False, copy=True, dtype=None,
                  compress="average", ret_sum=False):
   
    if not isinstance(normalize, bool):
        raise ValueError("`normalize` must be boolean!")
    if not isinstance(copy, bool):
        raise ValueError("`copy` must be boolean!")
    if not isinstance(ret_sum, bool):
        raise ValueError("`ret_sum` must be boolean!")
    if normalize and ret_sum:
        raise ValueError("'normalize' and 'ret_sum' must not both be True!")
    compress_values = ["average", "first", "second"]
    if compress not in compress_values:
        raise ValueError("Invalid value for `compress`! Possible values "
                         "are '{}'.".format(','.join(compress_values)))

    if dtype is None:
        dtype = np.dtype(a[0].__class__)
    else:
        dtype = np.dtype(dtype)

    ZERO_CUTOFF = 1e-15

    # If copy is false and dtype is the same as the input array,
    # then this line does not have an effect:
    trace = np.array(a, dtype=dtype, copy=copy)
    trace = trace.transpose()

    # Check parameters
    if m // 2 != m / 2:
        mold = m
        m = np.int_((m // 2 + 1) * 2)
    else:
        m = np.int_(m)

    N = N0 = trace.shape[0]

    # Find out the length of the correlation function.
    # The integer k defines how many times we can average over
    # two neighboring array elements in order to obtain an array of
    # length just larger than m.
    k = np.int_(np.floor(np.log2(N / m)))

    # In the base2 multiple-tau scheme, the length of the correlation
    # array is (only taking into account values that are computed from
    # traces that are just larger than m):
    lenG = m + k * (m // 2) + 1

    G = np.zeros((lenG, 2), dtype=dtype)

    normstat = np.zeros(lenG, dtype=dtype)
    normnump = np.zeros(lenG, dtype=dtype)

    traceavg = np.average(trace, axis = 1)

    # We use the fluctuation of the signal around the mean
    if normalize:
        trace -= traceavg.reshape(len(traceavg),1)

    # Otherwise the following for-loop will fail:
    if N < 2 * m:
        raise ValueError("`len(a)` must be >= `2m`!")

    # Calculate autocorrelation function for first m+1 bins
    # Discrete convolution of m elements
    for n in range(0, m + 1):
        G[n, 0] = deltat * n
        # This is the computationally intensive step
        G[n, 1] = np.sum(trace[:N - n] * trace[n:])
        normstat[n] = N - n
        normnump[n] = N
    # Now that we calculated the first m elements of G, let us
    # go on with the next m/2 elements.
    # Check if len(trace) is even:
    if N % 2 == 1:
        N -= 1
    # compress every second element
    if compress == compress_values[0]:
        trace = (trace[:N:2] + trace[1:N:2]) / 2
    elif compress == compress_values[1]:
        trace = trace[:N:2]
    elif compress == compress_values[2]:
        trace = trace[1:N:2]
    N //= 2
    # Start iteration for each m/2 values
    for step in range(1, k + 1):
        # Get the next m/2 values via correlation of the trace
        for n in range(1, m // 2 + 1):
            npmd2 = n + m // 2
            idx = m + n + (step - 1) * m // 2
            if len(trace[:N - npmd2]) == 0:
                # This is a shortcut that stops the iteration once the
                # length of the trace is too small to compute a corre-
                # lation. The actual length of the correlation function
                # does not only depend on k - We also must be able to
                # perform the sum with respect to k for all elements.
                # For small N, the sum over zero elements would be
                # computed here.
                #
                # One could make this for-loop go up to maxval, where
                #   maxval1 = int(m/2)
                #   maxval2 = int(N-m/2-1)
                #   maxval = min(maxval1, maxval2)
                # However, we then would also need to find out which
                # element in G is the last element...
                G = G[:idx - 1]
                normstat = normstat[:idx - 1]
                normnump = normnump[:idx - 1]
                # Note that this break only breaks out of the current
                # for loop. However, we are already in the last loop
                # of the step-for-loop. That is because we calculated
                # k in advance.
                break
            else:
                G[idx, 0] = deltat * npmd2 * 2**step
                # This is the computationally intensive step
                G[idx, 1] = np.sum(trace[:N - npmd2] *
                                   trace[npmd2:])
                normstat[idx] = N - npmd2
                normnump[idx] = N
        # Check if len(trace) is even:
        if N % 2 == 1:
            N -= 1
        # compress every second element
        if compress == compress_values[0]:
            trace = (trace[:N:2] + trace[1:N:2]) / 2
        elif compress == compress_values[1]:
            trace = trace[:N:2]
        elif compress == compress_values[2]:
            trace = trace[1:N:2]

        N //= 2

    if normalize:
        G[:, 1] /= traceavg**2 * normstat
    elif not ret_sum:
        G[:, 1] *= N0 / normnump

    if ret_sum:
        return G, normstat
    else:
        return G

    
def fcs2corr(data, dwell_time, list_of_g=['central', 'sum3', 'sum5', 'chessboard', 'ullr'], accuracy=50, algorithm='multipletau', averaging=None, list_of_g_out=None):
    """
    Convert SPAD-fcs data to correlation curves
    Function used for most of the correlations calculations.
    
    Parameters
    ----------
    data : np.array(Nt x Nc)
        Data variable, i.e. output from binfile2data.
        Nt: number of time points
        Nc: number of channels (typically 25 for a 5x5 detector)
    dwell_time : float
        Bin time [in µs].
    list_of_g : list, optional
        List of correlations to be calculated.
        The default is ['central', 'sum3', 'sum5', 'chessboard', 'ullr'].
    accuracy : int, optional
        Accuracy of the autocorrelation function. The default is 50.

    Returns
    -------
    G : object
        Object with all autocorrelations
        E.g. G.central contains the array with the central detector
        element autocorrelation.

    """
    
    # object from correlations class in which all correlation data is stored
    G = Correlations()
    
    # dwell time
    G.dwellTime = dwell_time

    if len(np.shape(data)) == 1:
        # vector is given instead of matrix, single detector only
        print('Calculating autocorrelation ')
        setattr(G, 'det0', correlate(data, data, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm))
        return G

    num_ch = np.shape(data)[1]
    for i in list_of_g:
        if isinstance(i, int):
            # autocorrelation of a detector element i
            print('Calculating autocorrelation of detector element ' + str(i))
            dataSingle = extract_spad_data(data, i)
            setattr(G, 'det' + str(i), correlate(dataSingle, dataSingle, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm))

        elif i == "central":
            # autocorrelation central detector element
            print('Calculating autocorrelation central detector element')
            dataCentral = extract_spad_data(data, "central")
            G.central = correlate(dataCentral, dataCentral, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)

        elif i[0] == 'x':
            # cross correlation two detector elements: e.g. 'x0112' for det1 x det12
            det0 = int(i[1:3])
            det1 = int(i[3:5])
            dataSingle0 = extract_spad_data(data, det0)
            dataSingle1 = extract_spad_data(data, det1)
            print('Calculating cross-correlation detector elements ' + str(det0) + 'x' + str(det1))
            Gtemp = correlate(dataSingle0, dataSingle1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            setattr(G, i, Gtemp)

        elif i == "sum3":
            # autocorrelation sum3x3
            print('Calculating autocorrelation sum3x3')
            dataSum3 = extract_spad_data(data, "sum3")
            G.sum3 = correlate(dataSum3, dataSum3, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)

        elif i == "sum5":
            # autocorrelation sum3x3
            print('Calculating autocorrelation sum5x5')
            dataSum5 = extract_spad_data(data, "sum5")
            G.sum5 = correlate(dataSum5, dataSum5, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
        elif i[0] == "C":
            # crosscorrelation custom sum of channels
            xpos = np.max([i.find('X'), i.find('x')])
            if xpos > -1:
                print('Calculating crosscorrelation custom sum')
                dataSum1 = extract_spad_data(data, i[0:xpos])
                dataSum2 = extract_spad_data(data, 'C' + i[xpos+1:])
            else:
                print('Calculating autocorrelation custom sum')
                dataSum1 = extract_spad_data(data, i)
                dataSum2 = dataSum1
            setattr(G, i, correlate(dataSum1, dataSum2, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm))
        
        elif i == "allbuthot":
            # autocorrelation sum5x5 except for the hot pixels
            print('Calculating autocorrelation allbuthot')
            dataAllbuthot = extract_spad_data(data, "allbuthot")
            G.allbuthot = correlate(dataAllbuthot, dataAllbuthot, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)

        elif i == "chessboard":
            # crosscorrelation chessboard
            print('Calculating crosscorrelation chessboard')
            dataChess0 = extract_spad_data(data, "chess0")
            dataChess1 = extract_spad_data(data, "chess1")
            G.chessboard = correlate(dataChess0, dataChess1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
        elif i == "chess3":
            # crosscorrelation small 3x3 chessboard
            print('Calculating crosscorrelation small chessboard')
            dataChess0 = extract_spad_data(data, "chess3a")
            dataChess1 = extract_spad_data(data, "chess3b")
            G.chess3 = correlate(dataChess0, dataChess1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)

        elif i == "ullr":
            # crosscorrelation upper left and lower right
            print('Calculating crosscorrelation upper left and lower right')
            dataUL = extract_spad_data(data, "upper_left")
            dataLR = extract_spad_data(data, "lower_right")
            G.ullr = correlate(dataUL, dataLR, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
        
        elif i == "twofocus":
            # crosscorrelations sum5left and sum5right, sum5top, and sum5bottom
            dataL = extract_spad_data(data, "sum5left")
            dataR = extract_spad_data(data, "sum5right")
            dataT = extract_spad_data(data, "sum5top")
            dataB = extract_spad_data(data, "sum5bottom")
            
            print('Calculating crosscorrelation two-focus left and right')
            G.twofocusLR = correlate(dataL, dataR, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusRL = correlate(dataR, dataL, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            print('Calculating crosscorrelation two-focus left and top')
            G.twofocusTL = correlate(dataT, dataL, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusLT = correlate(dataL, dataT, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            print('Calculating crosscorrelation two-focus left and bottom')
            G.twofocusLB = correlate(dataL, dataB, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusBL = correlate(dataB, dataL, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            print('Calculating crosscorrelation two-focus right and top')
            G.twofocusRT = correlate(dataR, dataT, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusTR = correlate(dataT, dataR, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            print('Calculating crosscorrelation two-focus right and bottom')
            G.twofocusRB = correlate(dataR, dataB, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusBR = correlate(dataB, dataR, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            print('Calculating crosscorrelation two-focus top and bottom')
            G.twofocusTB = correlate(dataT, dataB, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            G.twofocusBT = correlate(dataB, dataT, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
            
            
        elif i == "crossCenter":
            # crosscorrelation center element with L, R, T, B
            dataCenter = extract_spad_data(data, 12)
            for j in range(25):
                print('Calculating crosscorrelation central element with ' + str(j))
                data2 = extract_spad_data(data, j)
                Gtemp = correlate(dataCenter, data2, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                setattr(G, 'det12x' + str(j), Gtemp)
        
        elif i == 'singleElement':
            # autocorrelation single element
            # check first non empty channel
            j = -1
            ch1Found = False
            while not ch1Found:
                j += 1
                data1 = extract_spad_data(data, j)
                if np.sum(data1) > 10:
                    ch1Found = True
                    ch1 = j
            if ch1Found:
                print('Autocorrelation element ' + str(ch1))
                Gtemp = correlate(data1, data1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                G.auto = Gtemp
        
        elif i == "2MPD":
            # crosscorrelation two elements
            # check first non empty channel
            j = -1
            ch1Found = False
            ch2Found = False
            while not ch1Found:
                j += 1
                data1 = extract_spad_data(data, j)
                if np.sum(data1) > 10:
                    ch1Found = True
                    ch1 = j
            while not ch2Found and j < np.shape(data)[1] - 1:
                j += 1
                data2 = extract_spad_data(data, j)
                if np.sum(data2) > 10:
                    ch2Found = True
                    ch2 = j
            if ch1Found and not ch2Found:
                ch2 = ch1
                data2 = extract_spad_data(data, ch2)
                ch2Found = True
            if ch1Found and ch2Found:
                print('Cross correlation elements ' + str(ch1) + ' and ' + str(ch2))
                Gtemp = correlate(data1, data2, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                G.cross12 = Gtemp
                print('Cross correlation elements ' + str(ch2) + ' and ' + str(ch1))
                Gtemp = correlate(data2, data1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                G.cross21 = Gtemp
                print('Autocorrelation element ' + str(ch1))
                Gtemp = correlate(data1, data1, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                G.auto1 = Gtemp
                print('Autocorrelation element ' + str(ch2))
                Gtemp = correlate(data2, data2, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                G.auto2 = Gtemp
                
        elif i == "crossAll":
            # crosscorrelation every element with every other element
            if algorithm == "sparse_matrices":
                print("Calculating all crosscorrelations with sparse matrices algorithm")
                [Gall, Gtimes] = fcs_sparse(data, dwell_time*1e-6, m=accuracy)
                Gtemp = np.zeros((len(Gtimes), 2))
                Gtemp[:,0] = np.squeeze(Gtimes)
                if averaging is None:
                    for j in range(num_ch):
                        for k in range(num_ch):
                            Gtemp[:,1] = Gall[:,j,k]
                            setattr(G, 'det' + str(j) + 'x' + str(k), Gtemp)
                else:
                    # average over multiple cross-correlations
                    if averaging == 'default':
                        averaging = corrs2average_for_stics()
                        avs = [i[1] for i in averaging] # which correlations to average
                        els = [i[0] for i in averaging] # their new names
                    else:
                        avs = averaging
                        els = list_of_g_out
                    for el, av in enumerate(avs):
                        singleAv = [int(ch_nr) for ch_nr in re.findall(r'\d+', av)]
                        Nav = int(len(singleAv) / 2)
                        Gtemp = np.zeros((len(Gtimes), 2))
                        Gtemp[:,0] = np.squeeze(Gtimes)
                        for ind_av in range(Nav):
                            Gtemp[:,1] += Gall[:, singleAv[2*ind_av], singleAv[2*ind_av+1]]
                        Gtemp[:,1] /= Nav
                        setattr(G, els[el], Gtemp)
                        
            else:
                for j in range(num_ch):
                    data1 = extract_spad_data(data, j)
                    for k in range(num_ch):
                        data2 = extract_spad_data(data, k)
                        print('Calculating crosscorrelation det' + str(j) + ' and det' + str(k))
                        Gtemp = correlate(data1, data2, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                        setattr(G, 'det' + str(j) + 'x' + str(k), Gtemp)
        
        elif i == "autoSpatial":
            # number of time points
            Nt = np.size(data, 0)
            # detector size (5 for SPAD)
            N = int(np.round(np.sqrt(np.size(data, 1)-1)))
            # G size
            M = 2 * N - 1
            deltats = range(0, 1, 1) # in units of dwell times
            G.autoSpatial = np.zeros((M, M, len(deltats)))
            # normalization
            print("Calculating average image")
            avIm = np.mean(data, 0)
            # avInt = np.mean(avIm[0:N*N]) - can't be used since every pixel
            # has a different PSF amplitude!!
            # for j in range(np.size(data, 0)):
                # data[j, :] = data[j, :] - avIm
            avIm = np.resize(avIm[0:N*N], (N, N))
            # calculate autocorrelation
            k = 0
            for deltat in deltats:
                print("Calculating spatial autocorr delta t = " + str(deltat * dwell_time) + " µs")
                for j in range(Nt-deltat):
                    im1 = np.resize(data[j, 0:N*N], (N, N))
                    im1 = np.ndarray.astype(im1, 'int64')
                    im2 = np.resize(data[j + deltat, 0:N*N], (N, N))
                    im2 = np.ndarray.astype(im2, 'int64')
                    # G.autoSpatial[:,:,k] = G.autoSpatial[:,:,k] + ssig.correlate2d(im1, im2)
                    # calculate correlation between im1 and im2
                    for shifty in np.arange(-4, 5):
                        for shiftx in np.arange(-4, 5):
                            # go through all detector elements
                            n = 0  # number of overlapping detector elements
                            Gtemp = 0
                            for detx in np.arange(np.max((0, shiftx)), np.min((5, 5+shiftx))):
                                for dety in np.arange(np.max((0, shifty)), np.min((5, 5+shifty))):
                                    GtempUnNorm = im1[dety, detx] * im2[dety-shifty, detx-shiftx]
                                    GtempNorm = GtempUnNorm - avIm[dety, detx] * avIm[dety-shifty, detx-shiftx]
                                    GtempNorm /= avIm[dety, detx] * avIm[dety-shifty, detx-shiftx]
                                    Gtemp += GtempNorm
                                    n += 1
                            Gtemp /= n
                            G.autoSpatial[shifty+4,shiftx+4,k] += Gtemp
                G.autoSpatial[:,:,k] /= (Nt-deltat)
                k = k + 1

        elif i == "av":
            # average of all 25 individual autocorrelation curves
            for j in range(25):
                # autocorrelation of a detector element j
                print('Calculating autocorrelation of detector element ' + str(j))
                dataSingle = extract_spad_data(data, j)
                Gtemp = correlate(dataSingle, dataSingle, m=accuracy, deltat=dwell_time*1e-6, normalize=True, algorithm=algorithm)
                setattr(G, 'det' + str(j), Gtemp)
            Gav = Gtemp[:, 1]
            for j in range(24):
                Gav = np.add(Gav, getattr(G, 'det' + str(j))[:, 1])
            Gav = Gav / 25
            G.av = np.zeros([np.size(Gav, 0), 2])
            G.av[:, 0] = Gtemp[:, 0]
            G.av[:, 1] = Gav

    return G


def fcs2corrsplit(data, dwell_time, list_of_g=['central', 'sum3', 'sum5', 'chessboard', 'ullr'], accuracy=50, split=10):
    """
    Chunk SPAD-fcs trace into different parts and calculate correlation curves

    Parameters
    ----------
    data : np.array(Nt x Nc)
        Data variable, i.e. output from binfile2data.
        Nt: number of time points
        Nc: number of channels (typically 25 for a 5x5 detector)
    dwell_time : float
        Bin time [in µs].
    list_of_g : list, optional
        List of correlations to be calculated.
        The default is ['central', 'sum3', 'sum5', 'chessboard', 'ullr'].
    accuracy : int, optional
        Accuracy of the autocorrelation function. The default is 50.
    split : float, optional
        Number of traces to split the data into
        E.g. split=10 will divide a 60 second stream in 10 six second
        traces and calculate G for each individual trace. The default is 10.

    Returns
    -------
    G : object
        Object with all autocorrelations
        E.g. G.central contains the array with the central detector
        element autocorrelation..

    """
    
    if split == 1:
        G = fcs2corr(data, dwell_time, list_of_g, accuracy)
    else:
        G = Correlations()
        G.dwellTime = dwell_time
        N = int(np.size(data, 0))
        chunkSize = int(np.floor(N / split))
        for j in list_of_g:
            # --------------------- CALCULATE CORRELATION ---------------------
            print('Calculating correlation ' + str(j))
            i = 0
            for chunk in range(split):
                print('     Chunk ' + str(chunk+1) + ' --> ', end = '')
                # ------------------ CHUNK ------------------
                if data.ndim == 2:
                    dataSplit = data[i:i+chunkSize, :]
                else:
                    dataSplit = data[i:i+chunkSize]
                newList = [j]
                Gsplit = fcs2corr(dataSplit, dwell_time, newList, accuracy)
                GsplitList = list(Gsplit.__dict__.keys())
                for k in GsplitList:
                    if k.find('dwellTime') == -1:
                        setattr(G, k + '_chunk' + str(chunk), getattr(Gsplit, k))
                i += chunkSize
            # ---------- CALCULATE AVERAGE CORRELATION OF ALL CHUNKS ----------
            if j == '2MPD':
                avListBase = ['cross12', 'cross21', 'auto1', 'auto2']
                for avBase in avListBase:
                    avList = list(G.__dict__.keys())
                    avList = [i for i in avList if i.startswith(avBase + '_chunk')]
                    print('Calculating average correlation ' + avBase)
                    Gav = sum(getattr(G, i) for i in avList) / len(avList)
                    setattr(G, avBase + '_average', Gav)
                print('Calculating average cross correlation')
                G.cross_average = (G.cross12_average + G.cross21_average) / 2
            else:
                # Get list of "root" names, i.e. without "_chunk"
                Gfields = list(G.__dict__.keys())
                t = [Gfields[i].split("_chunk")[0] for i in range(len(Gfields))]
                t = list(dict.fromkeys(t))
                t.remove("dwellTime")
                # average over chunks
                for field in t:
                    print('Calculating average correlation ' + str(field))
                    avList = [i for i in Gfields if i.startswith(field + '_chunk')]
                    Gav = sum(getattr(G, i) for i in avList) / len(avList)
                    setattr(G, str(field) + '_average', Gav)
    return G


def fcs_load_and_corr_split(fname, list_of_g=['central', 'sum3', 'sum5', 'chessboard', 'ullr'], accuracy=16, split=10, time_trace=False, metadata=None, root=0, list_of_g_out=None, averaging=None, algorithm='multipletau'):
    """
    Load data from a file in chunks and calculate correlations.

    Parameters
    ----------
    fname : string
        File name.
    list_of_g : list, optional
        List of correlations to be calculated.
        The default is ['central', 'sum3', 'sum5', 'chessboard', 'ullr'].
    accuracy : int, optional
        Accuracy of the autocorrelation function. The default is 16.
    split : float, optional
        Number of traces to split the data into
        E.g. split=10 will divide a 60 second stream in 10 six second
        traces and calculate G for each individual trace. The default is 10.
    time_trace : boolean, optional
        see output. The default is False.
    metadata : None or object, optional
        if None: metadata is extracted from .txt info file
        if metadata Object: meta is given as input parameter. The default is None.
    root : int, optional
        used for GUI only to pass progress. The default is 0.
    list_of_g_out : TYPE, optional
        used for GUI only. The default is None.

    Returns
    -------
    G : object
        Object with all autocorrelations
        E.g. G.central contains the array with the central detector
        element autocorrelation.
    data : np.array()
        if time_trace == False: last chunk of raw data
        if time_trace == True: full time trace binned: [Nx25] array.

    """
    
    if fname.endswith(".czi"):
        fname_h5 = fname[:-4] + '.h5'
        if not os.path.exists(fname_h5):
            print('converting czi file')
            fname = czi2h5(fname)
    
    if metadata is None:
        metafile = get_metafile_from_file(fname)
        metadata = get_file_info(metafile)
    
    try:
        dwellTime = 1e-6 * metadata.timeResolution # s
    except:
        dwellTime = 1e-6 * metadata.time_resolution # s
    duration = metadata.duration
    
    N = int(np.floor(duration / split)) # number of chunks
    
    if N == 0:
        return [None, None]
    
    G = Correlations()
    G.dwellTime = dwellTime
    chunkSize = int(np.floor(split / dwellTime))
    for chunk in range(N):
        # --------------------- CALCULATE CORRELATIONS SINGLE CHUNK ---------------------
        if root != 0:
            root.progress = chunk / N
        string2print = "| Loading chunk " + str(chunk+1) + "/" + str(N) + " |"
        stringL = len(string2print) - 2
        print("+" + "-"*stringL + "+")
        print(string2print)
        print("+" + "-"*stringL + "+")
        data = file_to_fcs_count(fname, np.uint8, chunkSize, chunk*chunkSize)
        if time_trace == True:
            binSize = int(chunkSize / 1000 * N) # time trace binned to 1000 data points
            bdata = bindata_chunks(data, binSize)
            bdataL = len(bdata)
            num_ch = np.shape(bdata)[1]
            print("chunksize = " + str(chunkSize) + " - " + str(binSize))
            print("N = " + str(N))
            if chunk == 0:
                timetraceAll = np.zeros((1000, num_ch), dtype=int)
            timetraceAll[chunk*bdataL:(chunk+1)*bdataL, :] = bdata 
        
        indk = 0
        for ind, j in enumerate(list_of_g):
            print('     --> ' + str(j) + ": ", end = '')
            # ------------------ CHUNK ------------------
            newList = [j]
            Gsplit = fcs2corr(data, 1e6*dwellTime, newList, accuracy, averaging=averaging, list_of_g_out=list_of_g_out, algorithm=algorithm)
            GsplitList = list(Gsplit.__dict__.keys())
            for k in GsplitList:
                if k.find('dwellTime') == -1:
                    attrname = k
                    if list_of_g_out is not None:
                        attrname = list_of_g_out[indk]
                        indk += 1
                    if not np.isnan(getattr(Gsplit, k)).any():
                        setattr(G, attrname + '_chunk' + str(chunk), getattr(Gsplit, k))
    
    # ---------- CALCULATE AVERAGE CORRELATION OF ALL CHUNKS ----------
    print("Calculating average correlations")
    # Get list of "root" names, i.e. without "_chunk"
    Gfields = list(G.__dict__.keys())
    t = [Gfields[i].split("_chunk")[0] for i in range(len(Gfields))]
    t = list(dict.fromkeys(t))
    t.remove("dwellTime")
    print(t)
    # average over chunks
    for field in t:
        avList = [i for i in Gfields if i.startswith(field + '_chunk')]
        # check if all elements have same dimension
        Ntau = [len(getattr(G, i)) for i in avList]
        avList2 = [avList[i] for i in range(len(avList)) if Ntau[i] == Ntau[0]]
        
        Gtemp = getattr(G, avList2[0]) * 0
        GtempSquared = getattr(G, avList2[0])**2 * 0
        for chunk in avList2:
            Gtemp += getattr(G, chunk)
            GtempSquared += getattr(G, chunk)**2
        
        Gtemp /= len(avList2)
        Gstd = np.sqrt(np.clip(GtempSquared / len(avList2) - Gtemp**2, 0, None))
        
        Gtot = np.zeros((np.shape(Gtemp)[0], np.shape(Gtemp)[1] + 1))
        Gtot[:, 0:-1] = Gtemp # [time, average]
        Gtot[:, -1] = Gstd[:,1] # standard deviation
        
        setattr(G, str(field) + '_average', Gtot)   
    
    # average over same shifts in case of 'crossAll'
    # if 'crossAll' in list_of_g:
    #     print("Calculating spatially averaged correlations.")
    #     spatialCorr = np.zeros([9, 9, len(G.det0x0_average)])
    #     for shifty in np.arange(-4, 5):
    #         for shiftx in np.arange(-4, 5):
    #             avList = spad_shift_vector_crosscorr([shifty, shiftx])
    #             avList = [s + '_average' for s in avList]
    #             Gav = sum(getattr(G, i) for i in avList) / len(avList)
    #             spatialCorr[shifty+4, shiftx+4, :] = Gav[:,1]
    #     G.spatialCorr = spatialCorr

    if time_trace == True:
        data = timetraceAll

    return G, data


def correlate(data1, data2, m=50, deltat=1, normalize=True, algorithm='multipletau'):
    if algorithm == 'multipletau':
        G = multipletau.correlate(data1, data2, m=m, deltat=deltat, normalize=normalize)
    elif algorithm == 'wiener-khinchin':
        # time trace to correlation using fft
        G = tt2corr(data1, data2, m=m, macro_time=deltat)
    else:
        # algorithm is time-tag-to-correlation
        ind1 = np.where((data1 != 0))[0]
        ind2 = np.where((data2 != 0))[0]
        weights1 = data1[ind1]
        weights2 = data2[ind2]
        G = atimes_2_corr(ind1, ind2, accuracy=m, w0=weights1, w1=weights2, macroTime=deltat, taumax=ind1[-1] / 100)
    return G
        


def fcs_spatialcorrav(G, N=5, returnType='3Darray'):
    spatialCorr = np.zeros([2*N-1, 2*N-1, len(G.det0x0_average)])
    for shifty in np.arange(-(N-1), N):
        for shiftx in np.arange(-(N-1), N):
            avList = spad_shift_vector_crosscorr([shifty, shiftx], N)
            avList = [s + '_average' for s in avList]
            Gav = sum(getattr(G, i) for i in avList) / len(avList)
            if returnType == '3Darray':
                spatialCorr[shifty+N-1, shiftx+N-1, :] = Gav[:,1]
            else:
                setattr(G, 'shifty_' + str(shifty) + '_shiftx_' + str(shiftx), Gav)
    if returnType == '3Darray':
        G.spatialCorr = spatialCorr
    return G


def fcs_av_chunks(G, listOfChunks):
    """
    Average for each correlation mode the chunks given in listOfChunks.
    Used to calculate the average correlation for the good chunks only.

    Parameters
    ----------
    G : object
        Correlations object that contains all correlations.
    listOfChunks : list
        List with chunk numbers used for the calculation of the average
        e.g. [1, 3, 4, 7].

    Returns
    -------
    G : object
        Same object as input but with the additional fields
        G.det10_F0_averageX
        G.sum3_F0_averageX
        G.sum5_F0_averageX
        G.det10_F1_averageX
        ...
        where the averages are calculated over the listed chunks only.

    """
    
    listOfCorr = list(G.__dict__.keys())
    listOfCorr2 = []
    
    for corr in listOfCorr:
        if 'average' not in corr and 'chunk' in corr and 'chunksOff' not in corr:
            pos = corr.find('chunk')
            listOfCorr2.append(corr[0:pos])
    
    # remove duplicates
    listOfCorr2 = list(dict.fromkeys(listOfCorr2))
    
    for corr in listOfCorr2:
        Gtemp = getattr(G, corr + 'chunk0') * 0
        GtempSquared = getattr(G, corr + 'chunk0') * 0
        for chunk in listOfChunks:
            Gtemp += getattr(G, corr + 'chunk' + str(chunk))
            GtempSquared += getattr(G, corr + 'chunk' + str(chunk))**2
        
        Gtemp /= len(listOfChunks)
        Gstd = np.sqrt(np.clip(GtempSquared / len(listOfChunks) - Gtemp**2, 0, None))
        
        Gtot = np.zeros((np.shape(Gtemp)[0], np.shape(Gtemp)[1] + 1))
        Gtot[:, 0:-1] = Gtemp # [time, average]
        Gtot[:, -1] = Gstd[:,1] # standard deviation
        
        setattr(G, corr + 'averageX', Gtot)
    
    return G


def fcs_crosscenter_av(G, returnField='_averageX', returnObj = True):
    """
    Average pair-correlations between central pixel and other pixels that are
    located at the same distance from the center

    Parameters
    ----------
    G : TYPE
        Correlations object that (at least) contains all
        cross-correlations between central pixel and all other pixels:
        G.det12x12_average, G.det12x13_average, etc..
    returnField : string, optional
        name of the field containing the average. The default is '_averageX'.
    returnObj : boolean, optional
        return np.array() with average correlation or object. The default is True.

    Returns
    -------
    G : object or np.array()
        Same object as input but with the additional field
        G.crossCenterAv, which contains array of 6 columns, containing
        averaged cross-correlations between central pixel and pixels
        located at a distance of
            | 0 | 1 | sqrt(2) | 2 | sqrt(5) | sqrt(8) |.

    """
    
    try:
        tau = G.det12x12_average[:,0]
    except:
        tau = G.det12x12_chunk0[:,0]
    G.crossCenterAv = np.zeros((len(tau), 6))
    
    # average autocorrelation center element
    try:
        G.crossCenterAv[:,0] = getattr(G, 'det12x12' + returnField)[:,1]
    except:
        G.crossCenterAv[:,0] = getattr(G, 'det12x12' + '_average')[:,1]
    
    # average pair-correlations 4 elements located at distance 1 from center
    try:
        G.crossCenterAv[:,1] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + returnField)[:,1] for det in [7, 11, 13, 17]])), 1)
    except:
        G.crossCenterAv[:,1] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + '_average')[:,1] for det in [7, 11, 13, 17]])), 1)
    
    # average pair-correlations 4 elements located at distance sqrt(2) from center
    try:
        G.crossCenterAv[:,2] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + returnField)[:,1] for det in [6, 8, 16, 18]])), 1)
    except:
        G.crossCenterAv[:,2] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + '_average')[:,1] for det in [6, 8, 16, 18]])), 1)
    
    # average pair-correlation 4 elements located at distance 2 from center
    try:
        G.crossCenterAv[:,3] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + returnField)[:,1] for det in [2, 10, 14, 22]])), 1)
    except:
        G.crossCenterAv[:,3] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + '_average')[:,1] for det in [2, 10, 14, 22]])), 1)
    
    # average pair-correlation 8 elements located at distance sqrt(5) from center
    try:
        G.crossCenterAv[:,4] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + returnField)[:,1] for det in [1, 3, 5, 9, 15, 19, 21, 23]])), 1)
    except:
        G.crossCenterAv[:,4] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + '_average')[:,1] for det in [1, 3, 5, 9, 15, 19, 21, 23]])), 1)
    
    # average pair-correlation 4 elements located at distance sqrt(8) from center
    try:
        G.crossCenterAv[:,5] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + returnField)[:,1] for det in [0, 4, 20, 24]])), 1)
    except:
        G.crossCenterAv[:,5] = np.mean(np.transpose(np.array([getattr(G, 'det12x' + str(det) + '_average')[:,1] for det in [0, 4, 20, 24]])), 1)
    
    if returnObj:
        return G
    else:
        return G.crossCenterAv


def fcs_bin_to_csv_all(folderName=[], Glist=['central', 'sum3', 'sum5', 'chessboard', 'ullr'], split=10):
    # PARSE INPUT
    if folderName == []:
        folderName = os.getcwd()
    folderName = folderName.replace("\\", "/")
    folderName = Path(folderName)
    
    # CHECK BIN FILES
    allFiles = list_files(folderName, 'bin')
    
    # GO THROUGH EACH FILE
    for file in allFiles:
        fileName = ntpath.basename(file)
        print("File found: " + fileName)
        [G, data] = fcs_load_and_corr_split(file, Glist, 50, split)
        corr2csv(G, file[0:-4], [0, 0], 0)


def plot_fcs_correlations(G, plotList='all', limits=[0, -1], vector=[], pColors='auto', yscale='lin'):
    """
    Plot correlation curves

    Parameters
    ----------
    G : object
        Object with all autocorrelations
        Possible attributes:
            det*,
            central, sum3, sum5, allbuthot, chessboard, ullr, av
            autoSpatial,
            det12x*
            dwellTime (is not plotted).
    plotList : list or string, optional
        correlations to plot. The default is 'all'.
    limits : list with two elements, optional
        tau boundaries to plot. The default is [0, -1].
    vector : list, optional
        List of vectors to plot for crosscorr. The default is [].
    pColors : string, optional
        Plot colors. The default is 'auto'.
    yscale : string, optional
        linear or log y scale. The default is 'lin'.

    Returns
    -------
    figure.

    """

    spatialCorrList = ['autoSpatial']

    start = limits[0]
    stop = limits[1]

    # plotList contains all attributes of G that have to be plotted
    if plotList == 'all':
        plotList = list(G.__dict__.keys())
        
    # remove dwellTime from plotList
    if 'dwellTime' in plotList:
        plotList.remove('dwellTime')

    if 'av' in plotList:
        # remove all single detector element correlations
        plotListRemove = fnmatch.filter(plotList, 'det?')
        for elem in plotListRemove:
            plotList.remove(elem)
        plotListRemove = fnmatch.filter(plotList, 'det??')
        for elem in plotListRemove:
            plotList.remove(elem)
    
    if np.size(fnmatch.filter(plotList, 'det12x??')) > 10:
        # replace all individual cross-correlations by single crossCenter element
        plotListRemove = fnmatch.filter(plotList, 'det12x?')
        for elem in plotListRemove:
            plotList.remove(elem)
        plotListRemove = fnmatch.filter(plotList, 'det12x??')
        for elem in plotListRemove:
            plotList.remove(elem)
        plotList.append('crossCenter')
    
    if fnmatch.filter(plotList, '*_'):
        plotListStart = plotList[0]
        # plot chunks of data and average
        plotList = list(G.__dict__.keys())
        plotList.remove('dwellTime')
        plotList = [i for i in plotList if i.startswith(plotListStart)]
        

    # -------------------- Check for temporal correlations --------------------
    plotTempCorr = False
    for i in range(np.size(plotList)):
        if plotList[i] not in spatialCorrList:
            plotTempCorr = True
            break

    if plotTempCorr:
        leg = []  # figure legend
        h = plt.figure()
        plt.rcParams.update({'font.size': 15})
        maxy = 0
        miny = 0
        minx = 25e-9
        maxx = 10
        pColIndex = 0
        for i in plotList:

            #if i not in list(G.__dict__.keys()):
              #  break

            if i in ["central", "central_average", "sum3", "sum3_average", "sum5", "sum5_average", "allbuthot", "allbuthot_average", "chessboard", "chessboard_average", "chess3", "chess3_average", "ullr", "ullr_average", "av", "cross12_average", "cross21_average", "cross_average", "auto1_average", "auto2_average"]:
                # plot autocorrelation
                Gtemp = getattr(G, i)
                plt.plot(Gtemp[start:stop, 0], Gtemp[start:stop, 1], color=plot_colors(i), linewidth=1.3)
                maxy = np.max([maxy, np.max(Gtemp[start+1:stop, 1])])
                miny = np.min([miny, np.min(Gtemp[start+1:stop, 1])])
                minx = Gtemp[start, 0]
                maxx = Gtemp[stop, 0]
                leg.append(i)

            elif i == 'crossCenter':
                for j in range(25):
                    Gsingle = getattr(G, 'det12x' + str(j))
                    # plotColor = color_from_map(distance2detelements(12, j), 0, np.sqrt(8))
                    plt.plot(Gsingle[start:stop, 0], Gsingle[start:stop, 1], color=plot_colors(i))
                    maxy = np.max([maxy, np.max(Gsingle[start+1:stop, 1])])
                    miny = np.min([miny, np.min(Gtemp[start+1:stop, 1])])
                    leg.append(i + str(j))
            
            elif i == 'crossCenterAv':
                tau = G.det12x12_average[:,0]
                for j in range(6):
                    plt.plot(tau[start:stop], G.crossCenterAv[start:stop, j], color=plot_colors(j))
                miny = np.min(G.crossCenterAv[start+10:stop,:])
                maxy = np.max(G.crossCenterAv[start+1:stop,:])
                leg = ['$\Delta r = 0$', '$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']

            elif i != 'autoSpatial' and i != 'stics' and i != 'crossAll' and i != 'crossVector':
                # plot autocorr single detector element
                if pColors == 'auto':
                    plt.plot(getattr(G, i)[start:stop, 0], getattr(G, i)[start:stop, 1])
                else:
                    plt.plot(getattr(G, i)[start:stop, 0], getattr(G, i)[start:stop, 1], color=plot_colors(pColors[pColIndex]))
                    pColIndex += 1
                maxy = np.max([maxy, np.max(getattr(G, i)[start+1:stop, 1])])
                miny = np.min([miny, np.min(getattr(G, i)[start+1:stop, 1])])
                minx = getattr(G, i)[start, 0]
                maxx = getattr(G, i)[stop, 0]
                if '_average' in i:
                    iLeg = i[0:-8]
                else:
                    iLeg = i
                leg.append(iLeg)

        # figure lay-out
        plt.xscale('log')
        plt.xlabel('Temporal shift [s]')
        plt.ylabel('G')
        if yscale == 'log':
            plt.yscale('log')
        else:
            plt.yscale('linear')
        axes = plt.gca()
        axes.set_xlim([minx, maxx])
        axes.set_ylim([miny, maxy])
        if np.size(leg) > 0 and np.size(leg) < 10 and 'crossCenter' not in plotList:
            axes.legend(leg)
        plt.tight_layout()
        
        if 'crossCenter' in plotList:
            plot_cross_center_scheme()
            
    # -------------------- Check for spatial correlations --------------------
    if 'autoSpatial' in plotList:
        Gtemp = G.autoSpatial
        Gmax = np.max(Gtemp)
        xmax = (np.size(Gtemp, 0)) / 2
        extent = [-xmax, xmax, -xmax, xmax]
        for j in range(np.size(Gtemp, 2)):
            h = plt.figure()
            plt.imshow(Gtemp[:, :, j], extent=extent, vmin=0, vmax=Gmax)
            plt.title('delta_t = ' + str(G.dwellTime * j) + ' µs')
    
    if 'crossAll' in plotList:
        Gtemp = G.spatialCorr
        tau = G.det0x0_average[:,0]
        for vector in [[4, 4], [3, 4], [3, 3], [2, 4], [2, 3], [2, 2]]:
            plt.plot(tau, Gtemp[vector[0], vector[1], :])
        plt.legend(['[0, 0]', '[1, 0]', '[1, 1]', '[0, 2]', '[2, 1]', '[2, 2]'])
        plt.xscale('log')
        plt.xlabel('Temporal shift [s]')
        plt.ylabel('G')
        axes.set_ylim([miny, np.max(Gtemp[:,:,2:])])
#        Gtemp = G.spatialCorr
#        Gmax = np.sort(Gtemp.flatten())[-2] # second highest number
#        extent = [-4, 5, -4, 5]
#        for j in range(np.size(Gtemp, 2)):
#            h = plt.figure()
#            plt.imshow(Gtemp[:, :, j], extent=extent, vmin=0)
#            plt.title('delta_t = ' + str(G.dwellTime * j) + ' µs')
    
    if 'crossVector' in plotList:
        Gtemp = G.spatialCorr
        tau = G.det0x0_chunk0[:,0]
        for i in range(len(vector)):
            vectorI = vector[i]
            plt.plot(tau, Gtemp[4+vectorI[0], 4+vectorI[1], :], label='[' + str(vectorI[0]) + ', ' + str(vectorI[1]) + ']')
        plt.xscale('log')
        plt.xlabel('Temporal shift [s]')
        plt.ylabel('G')
        plt.legend()
        axes.set_ylim([miny, np.max(Gtemp[:,:,2:])])
    
    if 'stics' in plotList:
        Gtemp = getattr(G, 'det12x12')
        Gplot = np.zeros([9, 9])
        N = 10
        indArray = np.concatenate(([0], np.round(np.logspace(0, np.log10(len(Gtemp) - 1), N)).astype('int')))
        for i in range(N):
            # go through all lag times
            ind = np.round(indArray[i])
            print(ind)
            for yshift in np.arange(-4, 5):
                for xshift in np.arange(-4, 5):
                    # go through each shift vector
                    detDiff = -yshift * 5 - xshift
                    Gv = 0
                    nG = 0
                    for det1 in range(25):
                        [y, x] = coord(det1)
                        if x-xshift < 0 or x-xshift>4 or y-yshift < 0 or y-yshift > 4:
                            # don't do anything
                            pass
                        else:
                            det2 = det1 + detDiff
                            print('det1 = ' + str(det1) + ' and det2 = ' + str(det2))
                            if det2 >= 0 and det2 <= 24:
                                Gv += getattr(G, 'det' + str(det1) + 'x' + str(det2))[int(ind), 1]
                                nG += 1
                    Gplot[yshift+4, xshift+4] = Gv / nG
            if i == 0:
                plotMax = np.max(Gplot)
            xmax = 9 / 2
            extent = [-xmax, xmax, -xmax, xmax]
            h = plt.figure()
            plt.imshow(Gplot, extent=extent, vmin=0, vmax=plotMax)
            plt.title('delta_t = ' + str(int(G.dwellTime * ind)) + ' µs')
            
    return h


def plot_g_surf(G):
    N = np.size(G, 0)
    N = (N - 1) / 2
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    x = y = np.arange(-N, N + 1)
    X, Y = np.meshgrid(x, y)
    ax.plot_surface(X, Y, G)
    return fig


def plot_cross_center_scheme():
    detEl = range(25)
    distances = np.zeros(25)
    for i in detEl:
        distances[i] = distance2detelements(12, i)
    distances = np.resize(distances, (5, 5))
    plt.figure()
    plt.imshow(distances, 'viridis')
    plt.title('Color scheme cross-correlations')


def corrs2average_for_stics(det_size=5, pixelsOff=[]):
    """
    Return a list of cross-correlations with identical shifts between their
    channels

    Parameters
    ----------
    det_size : int, optional
        Number of pixels in each dimension of the array detector.
        The default is 5.
    pixelsOff : list, optional
        List of pixels that are excluded for this analysis because they are
        too far away from the optical axis, e.g [0,1,3,4,5,9,15,19,20,21,23,24]
        The default is [].

    Returns
    -------
    average : list
        List of pairs of strings describing the correlations to average and their
        new name, e.g.
        [['up', '12x0+12x1+12x2'], ['down', '0x12+1x12+2x12']]
    """
    listOfX = []
    listOfY = []
    names = []
    for vert in range(int(2*det_size-1)):
        for hor in range(int(2*det_size-1)):
            shifty = vert-det_size+1
            shiftx = hor-det_size+1
            listOfY.append(shifty)
            listOfX.append(shiftx)
            names.append('V' + str(shifty) + '_H' + str(shiftx))
    avList = [list_of_pixel_pairs_at_distance([listOfY[i], listOfX[i]], pixelsOff=pixelsOff) for i in range(len(listOfY))]
    avListStr = []
    names_final = []
    for idx, avSingleDist in enumerate(avList):
        if len(avSingleDist) > 0:
            avstr = ''
            for j in avSingleDist:
                avstr += str(j[0]) + 'x' + str(j[1]) + '+'
            avListStr.append(avstr[0:-1])
            names_final.append(names[idx])
        
    average = [[names_final[i], avListStr[i]] for i in range(len(names))]
    
    return average


import numpy as np


class BHspcInfo():
    pass


def read_spc(fname):
    """
    Read an asci BH fcs file and return the fcs photon arrival times

    Parameters
    ----------
    fname : string
        *asc file name
        To convert BH .spc file to ascii, use the Spcm software:
            - main > convert > FIFO files
            - setup file name: .set file
            - source file name: .spc file name
            - file format: ASCII with info header
            - extract photons data to ASCII file.

    Returns
    -------
    data : np.array
        Matrix containing the arrival times of all photons
        [units of clock time]
            - Column 1: photon arrival times, macrotimes [a.u.]
            - Column 2: photon arrival times, microtimes [a.u.]
            - Column 3: routing channel
            - Column 4: invalid flag.
    info : object
        Object containing the metadata
        - info.macrotime: macro time clock [s]
        - info.microtime: micro time unit [s]
        - info.fifotype: FIFO type
        - info.photons: total number of photons
        - info.invalidphotons: total number of invalid photons
        - info.fifooverruns: number of FIFO overruns.

    """
    
    # Open file
    print("Opening spc ASCII file.")
    with open(fname, mode='r') as file:
        rawdata = file.read()
    print("File opened.")
    
    # Variable to store metadata
    info = BHspcInfo()
    
    # Find macrotime
    MacroTimeStart = rawdata.find("Macro Time Clock [ns]", 0)
    MacroTimeStart = rawdata.find(": ", MacroTimeStart)
    MacroTimeStart += 2
    MacroTimeStop = rawdata.find(" ", MacroTimeStart)
    MacroTime = float(rawdata[MacroTimeStart:MacroTimeStop])
    info.macrotime = 1e-9 * MacroTime
    
    # Find microtime
    MicroTimeStart = rawdata.find("Micro Time Unit [ps]", 0)
    MicroTimeStart = rawdata.find(": ", MicroTimeStart)
    MicroTimeStart += 2
    MicroTimeStop = rawdata.find("\n", MicroTimeStart)
    MicroTime = float(rawdata[MicroTimeStart:MicroTimeStop])
    info.microtime = 1e-12 * MicroTime
    
    # Find FIFO time
    FifoTypeStart = rawdata.find("FIFO type ", 0)
    FifoTypeStart = rawdata.find(": ", FifoTypeStart)
    FifoTypeStart += 2
    FifoTypeStop = rawdata.find(" ", FifoTypeStart)
    FifoType = float(rawdata[FifoTypeStart:FifoTypeStop])
    info.fifotype = FifoType
    
    # Find number of photons
    NphotonsStart = rawdata.find("Total number of extracted photons", 0)
    NphotonsStart = rawdata.find(": ", NphotonsStart)
    NphotonsStart += 2
    NphotonsStop = rawdata.find("\n", NphotonsStart)
    Nphotons = rawdata[NphotonsStart:NphotonsStop]
    Nphotons = int(Nphotons)
    info.photons = Nphotons
    
    # Find number of invalid photons
    NphotonsInvStart = rawdata.find("invalid", NphotonsStop)
    NphotonsInvStart = rawdata.find(": ", NphotonsInvStart)
    NphotonsInvStart += 2
    NphotonsInvStop = rawdata.find("\n", NphotonsInvStart)
    NphotonsInv = rawdata[NphotonsInvStart:NphotonsInvStop]
    NphotonsInv = int(NphotonsInv)
    info.invalidphotons = NphotonsInv
    
    # Find number of FIFO overruns
    NfifoOverRunsStart = rawdata.find("number of fifo overruns", 0)
    NfifoOverRunsStart = rawdata.find(": ", NfifoOverRunsStart)
    NfifoOverRunsStart += 2
    NfifoOverRunsStop = rawdata.find("\n", NfifoOverRunsStart)
    NfifoOverRuns = rawdata[NfifoOverRunsStart:NfifoOverRunsStop]
    NfifoOverRuns = int(NfifoOverRuns)
    info.fifooverruns = NfifoOverRuns
    
    # Create empty array for photon arrival times
    data = np.zeros((Nphotons, 4), 'int64')
    
    # Check each line and extract photon arrival time
    # Find start position in the file
    start = rawdata.find("End of info header\n", 0)
    start = rawdata.find("\n", start)
    start += 1
    start = rawdata.find("\n", start)
    start += 1
    print(str(Nphotons) + " photons found.")
    for i in range(Nphotons):
        # macrotime
        stop = rawdata.find(" ", start)
        photonMacroTime = int(rawdata[start:stop])
        data[i, 0] = photonMacroTime
        # microtime
        start = stop + 1
        stop = rawdata.find(" ", start)
        photonMicroTime = int(rawdata[start:stop])
        data[i, 1] = photonMicroTime
        # routing channel
        start = stop + 2
        stop = rawdata.find(" ", start)
        photonChannel = int(rawdata[start:stop])
        data[i, 2] = photonChannel
        # invalid flag
        start = stop + 1
        stop = rawdata.find("\n", start)
        flag = int(rawdata[start:stop])
        data[i, 3] = flag
        # new line
        start = rawdata.find("\n", stop)

    return data, info



import csv
import numpy as np


def corr2csv(g, fname, limits=[0, 0], chunks=1):
    """
    Save all autocorrelations to separate .csv files for fit analysis in e.g.
    PyCorrFit.

    Parameters
    ----------
    g : object
        Autocorrelation data.
    fname : string
        filename base to store data (without file extension).
    limits : list of two numbers, optional
        crop data, indices of the start and stop index in lag time
    chunks : scalar, optional
        0: do not store fields with the word 'chunk' in it
        1: store all fields. The default is 1.

    Returns
    -------
    saved .csv files.

    """
    
    Glist = list(g.__dict__.keys())
    start = limits[0]
    stop = limits[1]
    for corr in Glist:
        if corr == 'dwellTime' or ('chunk' in corr and chunks==0):
            pass
        else:
            Gsingle = np.copy(getattr(g, corr))
            if len(np.shape(Gsingle)) == 2:
                # normal G(tau) data found
                pass
            else:
                # 3D array with cross-correlations found
                dim = np.shape(Gsingle)
                Greshaped = np.zeros([dim[2], dim[0]*dim[1]])
                for row in range(dim[0]):
                    for col in range(dim[1]):
                        Greshaped[:,row*dim[1]+col] = Gsingle[row, col, :]
                Gsingle = Greshaped
            if stop == 0:
                # stop = end
                stop = len(Gsingle)
            Gsingle = Gsingle[start:stop, :]
            # Gsingle[:, 0] = Gsingle[:, 0] * 1000  # convert time stamps to ms
            GsingleList = Gsingle.tolist()
    
            csv.register_dialect('myDialect',
                                 quoting=csv.QUOTE_MINIMAL,
                                 skipinitialspace=True)
    
            with open(fname + '_' + corr + '.csv', 'w', newline='') as f:
                writer = csv.writer(f) # dialect='myDialect'
                for row in GsingleList:
                    writer.writerow(row)
            f.close()



from .color_from_map import *
from .array2tiff import *
from .array2video import *
from .bindata import *
from .calcconc import *
from .calcdist_from_coord import *
from .cast_data import *
from .checkfname import *
from .closefile import *
from .cmap2d import *
from .cmaps import *
from .constants import *
from .csv2array import *
from .draw_ellipsoid import *
from .find_nearest import *
from .fit_data import *
from .fit_doughnut2d import *
from .fit_gauss_2d import *
from .fit_curve import *
from .fit_psf import *
from .goto import *
from .h5data import *
from .list_files import *
from .moving_average import *
from .np2hdf5 import *
from .path2fname import *
from .phasor import *
from .plot_colors import *
from .savefigure import *
from .savevar import *
from .stokes_einstein import *
from .time2string import *


from .checkfname import checkfname
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation


def array2video(data, fname='video.mp4', norm=False, interval=100, cmap='hot'):
    """
    Create video from 3D data array

    Parameters
    ----------
    data : np.array
        3D np.array [Ny x Nx x Nt].
    fname : string, optional
        File name to store the video. The default is 'video.mp4'.
    norm : boolean, optional
        Normalize the color map for each frame separately
        If false, the same color map is used for the whole stack. The default is False.
    interval : int, optional
        Number of ms per frame. The default is 100.

    Returns
    -------
    .mp4 file with the video.

    """
    
    # number of frames
    Nt = np.size(data, 2)
    
    # create empty variable to store data frames
    ims = []
    Imin = np.min(data)
    Imax = np.max(data)
    
    fig = plt.figure()
    
    FigSize = 10.5 # must be 10.5 to make the array size and video resolution match??
    
    fig.set_size_inches(FigSize * np.size(data, 0) / np.size(data, 1), FigSize, forward=False)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)
    for i in range(Nt):
        if norm:
            im = ax.imshow(data[:,:,i], cmap=cmap)
        else:
            im = ax.imshow(data[:,:,i], vmin=Imin, vmax=Imax, cmap=cmap)
        ims.append([im])
    
    ani = animation.ArtistAnimation(fig, ims, interval=interval, blit=True)
    
    fname = checkfname(fname, 'mp4')
    
    ani.save(fname)



def add_sum_column(data):
    """
    Add a column to a 2D array with the sum of all the other columns

    Parameters
    ----------
    data : 2D numpy array
        2D array with numbers.

    Returns
    -------
    data : 2D numpy array
        2D array with an additional column containing the sum of the 
        other columns

    """
    
    # import functions
    import numpy as np
    
    # data array size
    N = np.size(data, 0)
    M = np.size(data, 1)
    
    data2 = data # back-up
    data = np.zeros((int(N), M + 1), int)
    data[:, :-1] = data2
    data[:, M] = np.sum(data2, 1)
    
    return(data)


import numpy as np
import pandas as pd
import h5py
import os

from .fcs2arrivaltimes import ATimesData
from ..tools.list_files import list_files
from ..tools.closefile import closefile

import libttp.ttp as ttp


def load_atimes_data(fname, channels=21, sysclk_MHz=240):
    """
    Load multichannel arrival times data from .hdf5 file and store in data object
    for further processing. Used for data collected by the BrightEyes-TTM.

    Parameters
    ----------
    fname : string
        Filename [.hdf5]
        containing the tables ['det0', 'det1', 'det2', ..., 'det20']
        with each table containing the macro and microtimes.
    channels : int or list of int, optional
        Either number describing the number of channels (typically 21 or 25)
        or list of channels that have to be loaded, e.g. [15, 17, 18].
        The default is 21.
    sysclk_MHz : int, optional
        System clock frequency [MHz]. The default is 240.

    Returns
    -------
    data : object
        Data object with a field for each channel
        Each field contains a [Np x 2] np.array
        with Np the number of photons,
        column 1 the absolute macrotimes,
        column 2 the absolute microtimes.

    """
    
    if isinstance(channels, int):
        # total number of channels is given, e.g. 21
        Nchannels = channels
        channels = [str(x) for x in range(channels)]
    else:
        # individual channel numbers are given, e.g. [15, 17, 18]
        channels = [str(x) for x in channels]
        Nchannels = len(channels)
    
    data = ATimesData()
    
    if fname[-4:] == "hdf5":
        with h5py.File(fname, 'r') as f:
            for ch in channels:
                print('Loading channel ' + ch)
                setattr(data, 'det' + ch, f['det' + ch][()])
        f.close()
    elif fname[-2:] == "h5":
        calibDict=ttp.calculateCalibFromH5(filenameH5=fname, listChannel=range(0,Nchannels))
        data = ATimesData()
        for ch in channels:
            print('Loading channel ' + ch)
            df = ttp.applyCalibDict(fname, channel=int(ch), calibDict=calibDict)
            macrotime = df['cumulative_step'] / (1e6 * sysclk_MHz) * 1e12
            microtime = df['dt_' + ch]            
            setattr(data, 'det' + ch, np.transpose([macrotime, microtime]))
        closefile(fname)
    
    return data


def write_atimes_data(data, channels, fname):
    """
    Write multichannel arrival times data to .hdf5

    Parameters
    ----------
    data : object
        Arrival times data, i.e. output from load_atimes_data_pandas().
    channels : int or list of int
        Either number describing the number of channels (typically 21)
        or list of channels that have to be loaded, e.g. [15, 17, 18].
    fname : str
        Filename [.hdf5].

    Returns
    -------
    

    """
    
    if isinstance(channels, int):
        # total number of channels is given, e.g. 21
        channels = [str(x) for x in range(channels)]
    else:
        # individual channel numbers are given, e.g. [15, 17, 18]
        channels = [str(x) for x in channels]
    
    with h5py.File(fname, 'w') as f:
        for ch in range(21):
            f.create_dataset('det' + str(ch), data=getattr(data, 'det' + str(ch)))


def load_atimes_data_pandas(fname, chunksize=1000000, macro_freq=240e6):
    """
    Load multichannel arrival times data from h5 file and store in data object
    for further processing

    Parameters
    ----------
    fname : str
        Filename.
    chunksize : int, optional
        Number of rows to read in a single chunk. The default is 1000000.
    macro_freq : float, optional
        Conversion factor to go from relative macrotimes to absolute
        macrotimes. The default is 240e6.

    Returns
    -------
    data : object
        Data object with a field for each channel
        Each field contains a [Np x 2] np.array
        with Np the number of photons,
        column 1 the absolute macrotimes,
        column 2 the absolute microtimes.

    """
    
    # convert macrotime frequency to macrotime step size
    macroStep = 1 / macro_freq
    
    # read hdf file
    dataR = pd.read_hdf(fname, iterator=True, chunksize=chunksize)
    
    # number of data chunks to read
    Nchunks = int(np.ceil(len(dataR.coordinates) / chunksize))
    
    myIter = iter(dataR)
    
    chunk = 1
    for dataChunk in myIter:
        print('Loading data chunk ' + str(chunk) + '/' + str(Nchunks))
        if chunk == 1:
            # initialize data object
            data = ATimesData()
            listOfChannels = [name[12:] for name in dataChunk.columns if name.startswith('microtime_ch')]
            for chNr in listOfChannels:
                setattr(data, "det" + chNr, np.array([]))
        # go through each channel
        cumstep = dataChunk['cumulative_step']
        for chNr in listOfChannels:
            dataSingleCh = dataChunk['microtime_ch' + chNr]
            microtime = dataSingleCh[dataSingleCh.notna()]
            macrotime = macroStep * cumstep[dataSingleCh.notna()]
            dataSingleCh = np.transpose([macrotime, microtime])
            dataSingleChTot = getattr(data, "det" + chNr)
            setattr(data, "det" + chNr, np.vstack([dataSingleChTot, dataSingleCh]) if dataSingleChTot.size else dataSingleCh)
        chunk += 1
    
    return data


def load_atimes_data_newprot(filename, sysclk_MHz=240.0, laser_MHz= 80.0, nchannel = 26, kC4=43, channels=25):
    data_head, data_filename = os.path.split(filename)
    fname = data_filename[:-4]
    
    df = ttp.readNewProtocolFileToPandas(filenameIn=filename, CHANNELS=channels, reorder_channels=25)
    
    myReturn=ttp.convertFromPandasDataFrame(df, filenameOutputHDF5=data_head+fname+'.h5',
                    sysclk_MHz = sysclk_MHz,
                    laser_MHz=laser_MHz,
                    dwell_time_us=100.,
                    list_of_channels=np.arange(0,channels),
                    autoCalibration=True,
                    kC4=45.,
                    textInPlot=None,
                    compressionLevel=1,
                    makePlots=True,
                    ignorePixelLineFrame = False,
                    coincidence_analysis = True)
    
    fnameOut = myReturn['filenameH5']
    #h_main=pd.read_hdf(fnameOut, key="main")
    #h_valid_L = h_main[h_main["valid_tdc_L"]==1]
    data=load_atimes_data(fnameOut, channels=channels)
    data.macrotime = 1e-12
    data.microtime = 1e-12
    
    return data


def atimes_h5_to_hdf5(fname, chunksize=1000000, macro_freq=240e6, channels=21):
    """
    Load multichannel arrival times data from .h5 file, remove NaN and
    store as .hdf5 file

    Parameters
    ----------
    fname : str
        .h5 filename.
    chunksize : int, optional
        Number of rows to read in a single chunk. The default is 1000000.
    macro_freq : float, optional
        Conversion factor to go from relative macrotimes to absolute
        macrotimes. The default is 240e6.
    channels : int or list of int, optional
        Either number describing the number of channels (typically 21 or 25)
        or list of channels that have to be loaded, e.g. [15, 17, 18].
        The default is 21.

    Returns
    -------
    None.

    """
    
    data = load_atimes_data_pandas(fname, chunksize, macro_freq)
    write_atimes_data(data, channels, fname[:-3] + '.hdf5')


def atimes_raw_2_h5all(folder, sysclk_MHz=240, laser_MHz=40, n_ch=21, destination_folder=""):
    """
    Load multichannel arrival times data from raw data files and store as .h5
    files. Repeat this for all raw data files in a given folder.
    
    +------------------------------------------------------------------------+
    | > THIS IS USUALLY THE FIRST FUNCTION TO CAlL AFTER A TTM MEASUREMENT < |
    +------------------------------------------------------------------------+
    
    Note: this function expects the files to have the .ttm file extension
    Change this manually if not the case

    Parameters
    ----------
    folder : path
        Folder to look into.
    sysclk_MHz : float, optional
        System clock frequency [MHz]. The default is 240.
    laser_MHz : float, optional
        Laser clock frequency [MHz]. The default is 40.
    n_ch : int, optional
        Number of channels. The default is 21.
    destination_folder : path to folder, optional
        Location to store the .h5 file
        If left empty, the current working directory will be taken.
        The default is "".

    Returns
    -------
    Each file will be converted to a .h5 file.

    """
    
    files = list_files(directory=folder, filetype='ttm', substr=False)
    
    for file in files:
        atimes_raw_2_h5(file, sysclk_MHz=sysclk_MHz, laser_MHz=laser_MHz, Nch=n_ch, destinationFolder=destination_folder)
    print("All files converted.")


def atimes_raw_2_h5(fname, sysclk_MHz=240, laser_MHz=40, n_ch=21, destination_folder=""):
    """
    Load multichannel arrival times data from raw data file and store as .h5
    file.

    Parameters
    ----------
    fname : str
        Raw data file name.
    sysclk_MHz : float, optional
        System clock frequency [MHz]. The default is 240.
    laser_MHz : float, optional
        Laser clock frequency [MHz]. The default is 40.
    Nch : int, optional
        Number of channels. The default is 21.
    destination_folder : path to folder, optional
        Location to store the .h5 file
        If left empty, the current working directory will be taken.
        The default is "".

    Returns
    -------
    None.

    """
    
    
    list_of_channels=np.arange(0,n_ch)
    
    if destination_folder == "":
        destination_folder = os.getcwd()
    
    ttp.convertDataRAW(filenameToRead = fname,
                                sysclk_MHz = sysclk_MHz,
                                laser_MHz = laser_MHz,
                                list_of_channels = list_of_channels, # list of channel [0,1,2,3]
                                compressionLevel = 1,                # Compression HDF5 file
                                #metadata=metadataDict,              # If present append metadata to the HDF5
                                destinationFolder = "",              # If not selected the default output folder is filenameFolder/output/
                                ignorePixelLineFrame = True,         # Does not calculate x,y,frame
                                )
    print('Done.')


def h5_atimes_2_data(fname, n_ch, sysclk_MHz=240):
    """
    Load multichannel arrival times data from h5 file and store in data object
    for further processing - newer version

    Parameters
    ----------
    fname : str
        File name.
    Nch : int
        Number of channels.
    sysclk_MHz : float, optional
        System clock frequency [MHz]. The default is 240.

    Returns
    -------
    data : object
        Data object with a field for each channel
        Each field contains a [Np x 2] np.array
        with Np the number of photons,
        column 1 the absolute macrotimes,
        column 2 the absolute microtimes.

    """
    
    calibDict=ttp.calculateCalibFromH5(filenameH5=fname, listChannel=range(0,n_ch))
    data = ATimesData()
    for ch in range(n_ch):
        print('Loading channel ' + str(ch))
        df = ttp.applyCalibDict(fname, channel=ch, calibDict=calibDict)
        
        macrotime = df['cumulative_step'] / (sysclk_MHz * 1e6)
        microtime = df['dt_' + str(ch)]
        
        setattr(data, 'det' + str(ch), np.transpose([macrotime, microtime]))
    
    return data



import matplotlib.pyplot as plt
import numpy as np

def cmap2d(lifetimeIm, intensityIm, ltimeBound=['auto', 'auto'], intBound=['auto', 'auto'], cmap='rainbow', plotFig=False, equalize_lum=True):
    """
    Make FLIM image by combining lifetimeImage and intensityImage

    Parameters
    ----------
    lifetimeIm : np.array()
        2D array with lifetime values.
    intensityIm : np.array()
        2D array with intensity values.
    ltimeBound : list, optional
        Lower and upper bound of lifetime values. The default is ['auto', 'auto'].
    intBound : list, optional
        Lower and upper bound of intensity values. The default is ['auto', 'auto'].
    cmap : string, optional
        color map. The default is 'rainbow'.

    Returns
    -------
    outputIm : np.array()
        FLIM image.

    """
    
    
    Ny = np.shape(lifetimeIm)[0]
    Nx = np.shape(lifetimeIm)[1]
    
    if intBound[0] == 'auto':
        intBound[0] = np.min(intensityIm)
    if intBound[1] == 'auto':
        intBound[1] = np.max(intensityIm)
    if ltimeBound[0] == 'auto':
        ltimeBound[0] = np.min(lifetimeIm)
    if ltimeBound[1] == 'auto':
        ltimeBound[1] = np.max(lifetimeIm)
    
    intensityIm = np.clip(intensityIm, intBound[0], intBound[1])
    intensityIm = (intensityIm - intBound[0]) / (intBound[1] - intBound[0])
    lifetimeIm = np.clip(lifetimeIm, ltimeBound[0], ltimeBound[1]) 
        
    outputIm = np.zeros((Ny, Nx, 3))
    
    
    if cmap != 'seaborn-colorblind':
        cmap = plt.get_cmap(cmap)
        #N = np.size(cmap.colors, 0)
        N = cmap.N
        cmapArray = np.zeros((N, 3))
        for i in range(N):
            cmapArray[i, :] = cmap(i)[0:3]
        # equalize luminescence
        if equalize_lum:
            lmax = np.max((299*cmapArray[:,0] + 587*cmapArray[:,1] + 114*cmapArray[:,2]) / 1000)
            for i in range(N):
                lum = (299*cmapArray[i,0] + 587*cmapArray[i,1] + 114*cmapArray[i,2]) / 1000
                lumfactor = np.min((lmax / lum, 1/np.max(cmapArray[i, :])))
                cmapArray[i, :] *= lumfactor
        
        # get index of color map for lifetime
        idx = (np.floor((lifetimeIm - ltimeBound[0]) / (ltimeBound[1] - ltimeBound[0]) * N)).astype(int)
        idx = np.clip(idx, 0, N-1)
        for k in range(3):
            outputIm[:,:,k] = np.take(cmapArray[:,k], idx) * intensityIm
        
        # get rgb lifetime colors and scale with intensity
        # for k in range(3):
        #     outputIm[:, :, k] = cmap(idx / N)[k] * intensityIm[i, j] for k in range(3)]
        # for i in range(Ny):
        #     for j in range(Nx):
        #         #outputIm[i, j, :] = [cmap.colors[idx[i, j]][k] * intensityIm[i, j] for k in range(3)]
        #         outputIm[i, j, :] = [cmap(idx[i, j] / N)[k] * intensityIm[i, j] for k in range(3)]
    
    if plotFig:
        plt.figure()
        plt.imshow(outputIm)
    
    return outputIm
    



import numpy as np


def extract_spad_photon_streams(data, mode):
    """
    Extract data from specific detector elements from SPAD-fcs time-tagging data

    Parameters
    ----------
    data : object
        Object with for each detecor element a 2D array [t, m, F1, F2]
        - t: vector with macrotimes photon arrivals
        - m: microtimes
        - F1, F2, ... filter weights (based on microtimes).
    mode : string
        Either
        - "sum3": sum central 9 detector elements
        - "sum5": sum all detector elements
        TODO
        - a number i between -1 and -26: sum all elements except for i
        - "sum": sum over all detector element and all time points
        - "sum3_5D": sum central 9 detector elements in [z, y, x, t, c] dataset
        - "outer": sum 16 detector elements at the edge of the array
        - "chess0": sum all even numbered detector elements
        - "chess1": sum all odd numbered detector elements
        - "upper_left": sum the 12 upper left detector elements
        - "lower_right": sum the 12 lower right detector elements.

    Returns
    -------
    datasum : np.array()
        2D array with concatenated and sorted detector elements data.

    """
    
    if mode == 'sum3':
        # get list of detector fields
        listOfFields = [6, 7, 8, 11, 12, 13, 16, 17, 18] # [4, 5, 6, 9, 10, 11, 14, 15, 16]
        listOfFields  = ['det' + str(i) for i in listOfFields]
    
    elif mode == 'sum5':
        # get list of detector fields
        listOfFields = list(data.__dict__.keys())
        listOfFields = [i for i in listOfFields if i.startswith('det')]
        
    # concatenate all photon streams
    for j in range(len(listOfFields)):
        if j == 0:
            datasum = getattr(data, listOfFields[j])
        else:
            datasum = np.concatenate((datasum, getattr(data, listOfFields[j])))
    
    # sort all photon streams
    datasum = datasum[datasum[:,0].argsort()]
    
    # remove photon pairs with the same macrotime
    [dummy, ind] = np.unique(datasum[:,0], return_index=True)
    datasum = datasum[ind,:]
    
    return datasum



def cast_data(data, type='int64'):
    """
    Convert data array to type cast

    Parameters
    ----------
    data : np.array
        Data array.
    type : string, optional
        data type. The default is 'int64'.

    Returns
    -------
    data : np.array
        Same data array, but typecasted.

    """
    
    data = data.astype(type)
    return data



def u64_to_counts(number):
    """
    Convert an unsigned 64 bit number to a 25 element array containing the
    photon counts for each of the 25 detector elements during a single dwell
    time.

    Parameters
    ----------
    number : int
        Integer number <= 2^64 containing the photon counts for half of
        the detector elements for 1 bin.

    Returns
    -------
    counts : np.array
        List of 26 integer numbers with the photon counts for each of
        the 25 detector elements + sum. Zeros are inserted for detector
        elements that are not stored in the number.

    Example:
        u64_to_counts(13091560953979690842) will return
        [0, 0, 0, 0, 0, 0, 26, 23, 21, 10, 12, 26, 554, 52, 5, 11, 22, 0, 0, 0,...
             0, 0, 0, 0, 0, 759]
    """
    
    if number < 32:
        counts = [0] * 26
        return counts

    # Convert number to binary list of 64 digits
    number = list('{0:064b}'.format(int(number)))

    # Get time tag
    tt = number[-4:]
    tt = ''.join(tt)
    tt = int(tt, 2)
    del number[-4:]

    # Get cluster type
    clust = number[-1]
    clust = int(clust, 2)
    del number[-1]

    # Create list with 25 0's, one for each detector element
    counts = [0] * 26

    if clust == 0:
        # Data point contains counts of detector elements 0-5, 17-24

        # Detector elements 0-5 (4 bits / element)
        for i in range(6):
            counts[i], number = binary_to_counts(number, 4)

        # Detector element 17 (6 bits long)
        counts[17], number = binary_to_counts(number, 6)

        # Detector element 18 (5 bits long)
        counts[18], number = binary_to_counts(number, 5)

        # Detector elements 19-24 (4 bits / element)
        for i in range(6):
            counts[i+19], number = binary_to_counts(number, 4)

    elif clust == 1:
        # Data point contains counts of detector elements 6-16

        # Detector element 6 (5 bits long)
        counts[6], number = binary_to_counts(number, 5)

        # Detector element 7 (6 bits long)
        counts[7], number = binary_to_counts(number, 6)

        # Detector element 8 (5 bits long)
        counts[8], number = binary_to_counts(number, 5)

        # Detector element 9-10 (4 bits long)
        counts[9], number = binary_to_counts(number, 4)
        counts[10], number = binary_to_counts(number, 4)

        # Detector element 11 (6 bits long)
        counts[11], number = binary_to_counts(number, 6)

        # Detector element 12 (10 bits long)
        counts[12], number = binary_to_counts(number, 10)

        # Detector element 13 (6 bits long)
        counts[13], number = binary_to_counts(number, 6)

        # Detector element 14-15 (4 bits long)
        counts[14], number = binary_to_counts(number, 4)
        counts[15], number = binary_to_counts(number, 4)

        # Detector element 14-15 (5 bits long)
        counts[16], number = binary_to_counts(number, 5)

    counts[25] = sum(counts)

    return counts


def binary_to_counts(number, length):
    """
    Convert the last elements of a binary list to the corresponding photon
    decimal number and remove these elements from the list. Both the number
    and the new list are returned.

    Parameters
    ----------
    number : list
        List of 0's and 1's.
    length : int
        Integer number representing the number of bits that make up the
        last photon count number in the list.

    Returns
    -------
    counts : int
        Decimal representation of the binary string of length 'length',
        formed by the last elements in the list.
    number : list
        New binary list with the last 'length' elements removed.

    
    Example:
        binary_to_coun(['1', '0', '1', '1'], 3) will return (3, ['1']).
        3: the decimal representation of the last elements in the list (0,1,1)
        ['1']: list with the remaining elements of the original list
    """

    counts = int(''.join(number[-length:]), 2)
    del number[-length:]
    return counts, number



import numpy as np
from scipy.optimize import least_squares
import scipy.signal as scisig


def get_w0_from_fingerprint(y, fit_info, param, N=200, weights=1, lower_bounds=[-1e3, -1e3, 1e-3, 1e-3, 1e-3, 1e-3, 1e-3], upper_bounds = [1e3, 1e3, 1e4, 1e4, 1e3, 1e4, 1e4]):
    """
    Extract excitation and emission PSFs from SPAD finger print

    Parameters
    ----------
    y : np.array()
        Matrix with SPAD finger print.
    fit_info : np.array
        np.array boolean vector with always 7 elements
        1 for a fitted parameter, 0 for a fixed parameter
        order: [x0, y0, w0Ex, M, lambdaEmEx, pxsize, pxdist]
        E.g. to fit sigma and offset this becomes [0, 0, 0, 1, 1].
    param : list
        List with starting values for all parameters:
        order: [x0, y0, A, sigma, offset].
    N : int, optional
        Resolution of the calculation. The default is 200.
    weights : np.array or int, optional
        Matrix with weight values, 1 for unweighted fit. The default is 1.
    lower_bounds : np.array, optional
        lower bounds for the fit.
        The default is [-1e3, -1e3, 1e-3, 1e-3, 1e-3, 1e-3, 1e-3].
    upper_bounds : np.array, optional
        upper bounds for the fit.
        The default is [1e3, 1e3, 1e4, 1e4, 1e3, 1e4, 1e4].

    Returns
    -------
    fitresult : fitresult
        Result of the nonlinear least squares fit

    """
    
    fit_info = np.asarray(fit_info)
    param = np.asarray(param)
    lower_bounds = np.asarray(lower_bounds)
    upper_bounds = np.asarray(upper_bounds)
    
    fitparamStart = param[fit_info==1]
    fixedparam = param[fit_info==0]
    lower_bounds = lower_bounds[fit_info==1]
    upper_bounds = upper_bounds[fit_info==1]
    
    fitresult = least_squares(fitfun_w0_from_fingerprint, fitparamStart, args=(fixedparam, fit_info, y, N, weights), bounds=(lower_bounds, upper_bounds))
    
    return fitresult
    

def fitfun_w0_from_fingerprint(fitparamStart, fixedparam, fit_info, y, n_datapoints=200, weights=1):
    """
    Caculate residuals to fit excitation and emission PSFs from SPAD finger print

    Parameters
    ----------
    fitparamStart : list
        List with starting values for the fit parameters:
        order: [x0, y0, w0Ex, M, lambdaEmEx, pxsize, pxdist]
        with    x0, y0      position of the Gaussian center
                            in the image plane [µm]
                w0Ex        1/e^2 radius excitation PSF [µm]
                M           magnification of the system
                            typically 500
                lambdaEmEx  ratio excitation / emission wavelength
                pxsize      SPAD pixel size [µm], typically 50 µm
                pxdist      SPAD distance between pixels [µm],
                            typically 20 µm.
    fixedparam : list
        List with values for the fixed parameters.
    fit_info : list
        Boolean list of always 7 elements
        1   fit this parameters
        0   keep this parameter fixed.
    y : np.array
        Matrix with experimental function values.
    n_datapoints : int, optional
        Resolution of the calculation. The default is 200.
    weights : np.array, optional
        Matrix with fit weights, 1 for unweighted fit. The default is 1.

    Returns
    -------
    res : np.array
        Vector with residuals: f(param) - y.

    """
    
    # extract parameters
    param = np.float64(np.zeros(7))
    param[fit_info==1] = fitparamStart
    param[fit_info==0] = fixedparam
    
    Ny = 5
    Nx = 5
    
    Iairy = calc_fingerprint(param, n_datapoints)
    
    res = np.reshape(weights * (Iairy - y), [Nx*Ny, 1])
    res = np.ravel(res)

    return res


def calc_fingerprint(param, n_datapoints):
    """
    Caculate SPAD finger print

    Parameters
    ----------
    param : list
        List with all parameters:
        order: [x0, y0, w0Ex, M, lambdaEmEx, pxsize, pxdist]
        with    x0, y0      position of the Gaussian center
                            in the image plane [µm]
                w0Ex        1/e^2 radius excitation PSF [µm]
                M           magnification of the system,
                            typically 500
                lambdaEmEx  ratio emission / excitation wavelength
                pxsize      SPAD pixel size [µm], typically 50 µm
                pxdist      SPAD distance between pixels [µm],
                            typically 20 µm.
    n_datapoints : int
        Resolution of the calculation, typically 200.

    Returns
    -------
    Iairy : np.array
        Matrix with SPAD finger print.

    """
    
    x0 = 1e-6 * param[0]
    y0 = 1e-6 * param[1]
    w0Ex = 1e-6 * param[2]
    M = param[3]
    lambdaEmEx = param[4]
    pxsize = 1e-6 * param[5]
    pxdist = 1e-6 * param[6]

    print("Trying w0 = " + str(w0Ex))

    # calculate additional parameters
    w0EffEx = M * w0Ex
    w0EffEm = lambdaEmEx * w0EffEx
    
    # generate grid
    Ny = 5
    Nx = 5
    x = np.linspace(-5 * w0EffEx, 5 * w0EffEx, 2 * n_datapoints + 1)
    y = np.linspace(-5 * w0EffEx, 5 * w0EffEx, 2 * n_datapoints + 1)
    xv, yv = np.meshgrid(x, y)
    dx = x[1] - x[0]
    
    # calculate Gaussian
    Iex = np.exp(-2 * (xv**2 + yv**2) / w0EffEx**2)
    Iem = np.exp(-2 * ((xv - x0)**2 + (yv - y0)**2) / w0EffEm**2)
    
    Itot = scisig.convolve2d(Iex, Iem, mode='same')
    Itot /= np.sum(Itot)
    
    # add squares for SPAD elements
    detCoordinates = np.zeros((Nx * Ny, 4), dtype=int) # store coordinates of all detector elements
    cc = n_datapoints # index center
    for i in range(Nx):
        centerSingleX = cc + (i - np.floor(Nx/2)) * (pxsize + pxdist) / dx
        Lx = int(centerSingleX - pxsize / dx / 2 + 1) # x coordinate left border
        Rx = int(centerSingleX + pxsize / dx / 2 + 1) # x coordinate right border
        for j in range(Ny):
            centerSingleY = cc + (j - np.floor(Ny/2)) * (pxsize + pxdist) / dx
            Ty = int(centerSingleY - pxsize / dx / 2 + 1) # y coordinate top border
            By = int(centerSingleY + pxsize / dx / 2 + 1) # y coordinate bottom border
            detCoordinates[i*Ny + j] = [Lx, Rx, Ty, By]
    
    Iairy = np.zeros((Ny, Nx))
    for i in range(Nx):
        for j in range(Ny):
            [Lx, Rx, Ty, By] = detCoordinates[i*Ny + j]
            Iairy[j, i] = np.sum((Itot[Ty:By, Lx:Rx]))
    
    # normalize
    Iairy /= np.max(Iairy)
    
    return Iairy


import numpy as np


def lifetimehist(data, m_bins=260, laser_freq=80e6):
    """
    Make lifetime histograms of all channels

    Parameters
    ----------
    data : object
        data object with macro and microtimes.
    m_bins : int, optional
        number of microtime bins. The default is 260.
    laser_freq : float, optional
        laser frequency (Hz). The default is 80e6.

    Returns
    -------
    data : object
        same data object as input, but with histograms added.

    """
    
    # get list of detector fields
    listOfFields = list(data.__dict__.keys())
    listOfFields = [i for i in listOfFields if i.startswith('det')]
    Ndet = len(listOfFields)
    
    for det in range(Ndet):
        
        # macrotimes
        macroTime = getattr(data, "det" + str(det))[:,0] # ps
        
        # calculate proper microtimes
        microTime = getattr(data, "det" + str(det))[:,1]
        microTime = np.mod(microTime, 1 / data.microtime / laser_freq)
        microTime = -microTime + np.max(microTime)
        
        # make histogram of microtimes
        [Ihist, lifetimeBins] = np.histogram(microTime, m_bins)
        lifetimeBins = lifetimeBins[0:-1]
        
        # store histogram
        setattr(data, "hist" + str(det), np.transpose(np.stack((lifetimeBins, Ihist))))
        setattr(data, "det" + str(det), np.transpose([macroTime, microTime]))
        
    return data



from ..tools.fit_gauss_2d import fit_gauss_2d
from ..tools.fit_curve import fit_curve
import numpy as np

def fcs2imsd(G, tau, fitparam=[1, 1, 0, 0], startvalues=[1,1,1,1], lbounds=[-1e9, -1e9, -1e9, -1e9], ubounds=[1e9, 1e9, 1e9, 1e9], remove_outliers=False, remove_afterpulsing=True):
    # G is 3D array with G(tau, xi, psi)
    # param = [D, offset, smoothing, pixel size]
        
    smoothing = int(startvalues[2])
    newlen = int(len(tau)//smoothing)
    var = np.zeros((newlen))
    taunew = np.zeros((newlen))
    
    # convert D to slope
    D = startvalues[0]
    rho = 1e-3 * startvalues[3] # µm
    slope = 2 / rho**2 * D
    startvalues[0] = slope
    
    # find sigma as a function of tau
    for i in range(int(len(tau)//smoothing)):
        # [x0, y0, A, sigma, offset]
        Gsingle = np.sum(G[i*smoothing:(i+1)*smoothing,:,:],0) / smoothing
        # remove central afterpulsing peak
        weights = 1
        if remove_afterpulsing:
            weights = np.ones(Gsingle.shape)
            weights[4, 4] = 0
        fitres = fit_gauss_2d(Gsingle, [1,1,1,1,1], [1,1,np.max((G[i,1,1]+1e-5, 0)),1,0])
        var[i] = fitres.x[3]**2
        taunew[i] = np.mean(tau[i*smoothing:(i+1)*smoothing])
    
    # remove outliers
    if remove_outliers:
        median_var = np.median(var)
        mask = var < 3 * median_var
        var = var[mask]
        taunew = taunew[mask]
        
    # fit linear curve
    fitres = fit_curve(var, taunew, 'linear', fitparam[0:2], startvalues[0:2], lbounds[0:2], ubounds[0:2], savefig=0)
    
    fitresult = startvalues
    fitresult[fitparam] = fitres.x
    
    # convert slope to D
    fitresult[0] *= (1e-3*startvalues[-1])**2 / 2
    
    return var, taunew, fitresult


import numpy as np
from scipy.optimize import least_squares
import matplotlib.pyplot as plt
from .fcs_analytical import fcs_analytical, fcs_dualfocus, fcs_2c_analytical, fcs_2c_2d_analytical, fcs_analytical_2c_anomalous, nanosecond_fcs_analytical, uncoupled_reaction_diffusion
from .plot_pycorrfit import plot_pycorrfit, PyCorrfit_data
from .mem_fit import mem_fit_free_diffusion


def fcs_fit(Gexp, tau, fitfun, fit_info, param, lBounds, uBounds, plotInfo, splitData=0, savefig=0, plotTau=True, weights=1):
    """
    Fit experimental fcs data to the analytical model
    Assuming 3D diffusion in a Gaussian focal volume
    No triplet state assumed

    Parameters
    ----------
    Gexp : 1D numpy array
        Vector with experimental autocorrelation curve (G).
    tau : 1D numpy array
        Vector with lag times [s].
    fitfun : string or function
        Fit function (either string or the actual function)
        'fitfun_2c'           Fit two components
        'fitfun_dualfocus'    Fit two-focus fcs
        'fitfun_circfcs'      Circular scanning-fcs.
        'fitfun_free_diffusion_2d'
        'fcs_analytical_2c_anomalous'
    fit_info : 1D numpy array
        np.array boolean vector with [N, tauD, SP, offset, 1e6*A, B, vx, vy,...]
        1 if this value has to be fitted
        0 if this value is fixed during the fit.
    param : 1D numpy array
        np.array vector with start values for [N, tauD, SP, offset, A, B, vx, vy]
        A ~ 1e6*2.5e-8
        B ~ -1.05.
    lBounds : 1D numpy array
        Lower bounds for the fit parameters.
    uBounds : 1D numpy array
        Upper bounds for the fit parameters.
    plotInfo : string
        "central", "sum3", or "sum5". Determines only the color of the plot
        -1 to not plot the figure.
    splitData : scalar, optional
        The default is 0.
    savefig : scalar or string, optional
        0 to not plot but not save the figure
        file name with extension to save as "png" or "eps". The default is 0.
    plotTau : TYPE, optional
        DESCRIPTION. The default is True.
    weights : TYPE, optional
        DESCRIPTION. The default is 1.

    Returns
    -------
    fitresult : object
        Output of least_squares
        fitresult.x = fit results
            tauD in [ms].

    """
    
    if fitfun != 'mem_fit_free_diffusion' and fitfun != mem_fit_free_diffusion:
        fitparamStart = param[fit_info==1]
        fixedparam = param[fit_info==0]
        lowerBounds = lBounds[fit_info==1]
        upperBounds = uBounds[fit_info==1]
    
    if type(fitfun) == str:
        if fitfun == 'fitfun_2c':
            fitresult = least_squares(fitfun_2c, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))
        if fitfun == 'fitfun_an':
            fitresult = least_squares(fitfun_an, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))
        elif fitfun == 'fitfun_dualfocus':
            fitresult = least_squares(fitfun_dualfocus, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, splitData, weights), bounds=(lowerBounds, upperBounds))
        elif fitfun == 'fitfun_circfcs':
            fitresult = least_squares(fitfun_circfcs, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))
        elif fitfun == 'mem_fit_free_diffusion':
            fitresult = mem_fit_free_diffusion(param, tau, Gexp, weights)
        elif fitfun == 'fitfun_free_diffusion_2d':
            fitresult = least_squares(fitfun_free_diffusion_2d, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))
        elif fitfun == 'fcs_analytical_2c_anomalous':
            fitresult = least_squares(fcs_analytical_2c_anomalous, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))
    else:
        if fitfun == fitfun_dualfocus:
            fitresult = least_squares(fitfun_dualfocus, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, splitData, weights), bounds=(lowerBounds, upperBounds))
        elif fitfun == mem_fit_free_diffusion:
            fitresult = mem_fit_free_diffusion(param, tau, Gexp, weights)
        elif fitfun == fcs_fit_dualfocus:
            fitresult = fcs_fit_dualfocus(Gexp, tau, fit_info, param, useSingleW=True, weights=weights)
        else:
            fitresult = least_squares(fitfun, fitparamStart, args=(fixedparam, fit_info, tau, Gexp, weights), bounds=(lowerBounds, upperBounds))

    fitresult.fun /= weights

    if plotInfo != -1:
        plot_fit(tau, Gexp, param, fit_info, fitresult, plotInfo, savefig, plotTau)
    
    return fitresult


def fcs_fit_dualfocus(Gexp, tau, fit_info, param, plotResults=False, useSingleW=False, weights=1):
    """
    Fit experimental fcs data to the analytical model
    Assuming 3D diffusion in a Gaussian focal volume
    No triplet state assumed

    Parameters
    ----------
    Gexp : 2D np.array
        2D matrix with correlation curves
        Each column is a correlation curve.
    tau : 1D np.array
        Vector with tau values for a single autocorrelation curve [s].
    fit_info : 1D np.array
        np.array boolean vector with
        [c, D, w0, ..., wN, SF0, ..., SFN, rhox0, ..., rhoxN, rhoy0, ..., rhoyN, vx, vy, dc0, ..., dcN]
        1 if this value has to be fitted
        0 if this value is fixed during the fit.
    param : 1D np.array
        vector with start values for all the parameters.
    plotResults : boolean, optional
        Plot results. The default is True.
    useSingleW : boolean, optional
        Use a single w0 value for all curves. The default is False.

    Returns
    -------
    fitresult : object
        Output of least_squares.

    """

    fitparamStart = param[fit_info==1]
    fixedparam = param[fit_info==0]
    
    # concatenate all G vectors in single array
    if len(np.shape(Gexp)) == 1:
        # Gexp is vector: only one G given
        NG = 1
    else:
        NG = np.size(Gexp, 1)
    Ntau = np.size(Gexp, 0)
    Gall = np.reshape(np.transpose(Gexp), [Ntau * NG])
    tauall = np.concatenate([tau for i in range(NG)])
    # create vector with ranges of individual correlations
    splitData = [i*Ntau for i in range(NG+1)]
    
    fitresult = least_squares(fitfun_dualfocus, fitparamStart, args=(fixedparam, fit_info, tauall, Gall, splitData, useSingleW, weights))
    
    # plot fit result
    if plotResults:
        Dfitted = param[2]**2 / 4 / (1e-3 * fitresult.x[1])
        residuals = fitresult.fun
        Gfitresult = Gall - residuals
        plt.figure()
        for i in range(NG):
            plt.plot(tau, Gfitresult[splitData[i]:splitData[i+1]])
        plt.xscale('log')
        plt.xlabel('tau (s)')
        plt.ylabel('G')
        plt.ylim([np.min(Gall), np.max(Gall)])
        plt.title('Fit results - Fitted D = ' + "{:.2e}".format(1e12*Dfitted) + ' µm^2/s')
        if NG == 6:
            leg = ['$\Delta r = 0$', '$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        elif NG == 5:
            leg = ['$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        plt.tight_layout()
        
        # plot fit residuals
        residuals = fitresult.fun
        plt.figure()
        for i in range(NG):
            plt.plot(tau, residuals[splitData[i]:splitData[i+1]])
        plt.xscale('log')
        plt.xlabel('tau (s)')
        plt.ylabel('G')
        plt.title('Fit residuals')
        if NG == 6:
            leg = ['$\Delta r = 0$', '$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        elif NG == 5:
            leg = ['$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        plt.tight_layout()
        
        # plot original G
        plt.figure()
        for i in range(NG):
            plt.plot(tau, Gall[splitData[i]:splitData[i+1]])
        plt.xscale('log')
        plt.xlabel('tau (s)')
        plt.ylabel('G')
        plt.ylim([np.min(Gall), np.max(Gall)])
        plt.title('Experimental cross-correlations')
        if NG == 6:
            leg = ['$\Delta r = 0$', '$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        elif NG == 5:
            leg = ['$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        plt.tight_layout()
        
        # plot original G and fit
        plt.figure()
        for i in range(NG):
            plt.plot(tau, Gall[splitData[i]:splitData[i+1]])
            plt.plot(tau, Gfitresult[splitData[i]:splitData[i+1]], 'k')
        plt.xscale('log')
        plt.xlabel('tau (s)')
        plt.ylabel('G')
        plt.ylim([np.min(Gall), np.max(Gall)])
        if NG == 6:
            leg = ['$\Delta r = 0$', '$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        elif NG == 5:
            leg = ['$\Delta r = 1$', '$\Delta r = \sqrt{2}$', '$\Delta r = 2$', '$\Delta r = \sqrt{5}$', '$\Delta r = 2\sqrt{2}$']
            plt.legend(leg)
        plt.tight_layout()

    return fitresult


def fitfun(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    fcs free diffusion fit function
    
    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD, SP, offset, A, B]
        E.g. if only N and tauD are fitted, this becomes a two
        element vector [1, 1e-3].
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD, SP, offset, 1e6*A, B]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 6 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0, 0]
        order: [N, tauD, SP, offset, 1e6*A, B].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Vector with weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(6))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    N = fitparam[0]
    tauD = fitparam[1]
    SF = fitparam[2]
    offset = fitparam[3]
    A = fitparam[4]
    B = -fitparam[5]

    # calculate theoretical autocorrelation function    
    FCStheo = fcs_analytical(tau, N, tauD, SF, offset, 1e-6*A, B)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_an(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    fcs fit anomalous diffusion function

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD, SP, offset, alpha]
        E.g. if only N and tauD are fitted, this becomes a two
        element vector [1, 1e-3].
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD, SP, offset, alpha]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 5 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0]
        order: [N, tauD, SP, offset, alpha].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Vector with weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(5))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    N = fitparam[0]
    tauD = fitparam[1]
    SF = fitparam[2]
    offset = fitparam[3]
    alpha = fitparam[4]

    # calculate theoretical autocorrelation function    
    FCStheo = fcs_analytical(tau, N, tauD, SF, offset, 0, 0, alpha)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res

def fitfun_an_2c(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    fcs fit anomalous diffusion function

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD1, tauD2, alpha1, alpha2, F, T, tautrip, SP, offset]
        E.g. if only N and tauD are fitted, this becomes a two
        element vector [1, 1e-3].
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD, SP, offset, alpha]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 5 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0]
        order: [N, tauD, SP, offset, alpha].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Vector with weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(10))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    N = fitparam[0]
    tauD1 = fitparam[1]
    tauD2 = fitparam[2]
    alpha1 = fitparam[3]
    alpha2 = fitparam[4]
    F = fitparam[5]
    T =fitparam[6]
    tau_triplet = fitparam[7]
    SF = fitparam[8]
    offset = fitparam[9]
    
    # calculate theoretical autocorrelation function    
    FCStheo = fcs_analytical_2c_anomalous(tau, N, tauD1, tauD2, alpha1, alpha2, F, T, tau_triplet, SF, offset)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res

def fitfun_2c(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    fcs fit function two free components

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, SF, offset, A, B]
        E.g. if only N and tauD1 are fitted, this becomes a two
        element vector [1, 1e-3].
        parameters  N       number of particles in focal volume [dim.less]
                    tauD1   diffusion time species 1 [ms]
                    tauD2   diffusion time species 2 [ms]
                    F       fraction of species 1 [dim.less]
                    alpha   relative molecular brightness [dim.less]
                    T       fraction in triplet [dim.less]
                    tautrip residence time in triplet state [µs]
                    SF      shape factor [dim.less]
                    offset  [dim.less]
                    A, B    afterpulsing properties
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, SF, offset, A, B]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 11 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, SF, offset, A, B].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(11))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    # convert to SI
    N = fitparam[0]
    tauD1 = 1e-3 * fitparam[1] # ms to s
    tauD2 = 1e-3 * fitparam[2] # ms to s
    F = fitparam[3]
    alpha = fitparam[4]
    T = fitparam[5]
    tautrip = 1e-6 * fitparam[6] # s
    SF = fitparam[7]
    offset = fitparam[8]
    A = 1e-6 * fitparam[9]
    B = -fitparam[10]

    # calculate theoretical autocorrelation function
    FCStheo = fcs_2c_analytical(tau, N, tauD1, tauD2, F, alpha, T, tautrip, SF, offset, A, B)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_dualfocus(fitparamStart, fixedparam, fit_info, tau, yexp, splitData, useSingleW=False, weights=1):
    """
    Dual focus fcs fit function

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [c, D, w2 for all, SF for all, rhox for all, rhoy for all, vx, vy, offset for all]
        E.g. if only c and D are fitted, this becomes a two
        element vector [1, 1e-3].
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD, SF, offset]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 4 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0]
        order: [N, tauD, SF, offset].
    tau : 1D np.array
        Vector with tau values
        All curves are concatenated.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    splitData : 1D np.array
        Vector with indices to split data
        has to start with 0 and end with N (= number of data points)
        E.g. [0 100 200] will split the data into two traces
        One from 0 to 99 and one from 100 to 199.
    useSingleW : 1D np.array, optional
        Use the same w0 value for all cross-correlation curves
        If True, "fitparamStart", "fixedparam" and "fit_info"
        remain as before, bu only the first w0 value is used.
        The default is False.
    weights : 1D np.array, optional
        Vector with weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Fit residuals.

    """
    
    sD = splitData
    Ntraces = len(sD) - 1
    
    fitparam = np.float64(np.zeros(4 + 5*Ntraces))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    N0 = fitparam[0]
    D = 1e-12*fitparam[1] # m^2/s
    #tauD0 = 1e-3 * fitparam[1] # ms -> s
    w = fitparam[2:2+Ntraces]
    if useSingleW:
        w = w * 0 + fitparam[2]
    
    w = 1e-9 * w # m
    SF = fitparam[2+Ntraces:2+2*Ntraces]
    rhox = 1e-9*fitparam[2+2*Ntraces:2+3*Ntraces] # m
    rhoy = 1e-9*fitparam[2+3*Ntraces:2+4*Ntraces] # m
    vx = 1e-9*fitparam[2+4*Ntraces:3+4*Ntraces] # m/s
    vy = 1e-9*fitparam[3+4*Ntraces:4+4*Ntraces] # m/s
    dc = fitparam[4+4*Ntraces:4+5*Ntraces]
    
    # get particle concentration
    c = N0 / np.pi**(3/2) / w[0]**3 / SF[0]
    # calculate N for all curves
    N = [np.pi**(3/2) * c * w[i]**3 * SF[i] for i in range(Ntraces)]
    
    # get diffusion coefficient
    #D = w[0]**2 / 4 / tauD0
    
    yModel = np.concatenate([fcs_dualfocus(tau[sD[i]:sD[i+1]], N[i], D, w[i], SF[i], rhox[i], rhoy[i], dc[i], vx, vy) for i in range(Ntraces)])

    res = yexp - yModel
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_circfcs(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    Dual focus fcs fit function

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD, w, SF, Rcirc, Tcirc, offset]
        E.g. if only N and tauD are fitted, this becomes a two
        element vector [1, 1e-3].
    fixedparam : 1D np.array
        List with values for the fixed parameters
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 7 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0, 0, 0].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        DESCRIPTION. The default is 1.

    Returns
    -------
    res : 1D np.array
        Fit residuals.

    """
    
    fitparam = np.float64(np.zeros(7))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    N = fitparam[0]
    tauD0 = 1e-3 * fitparam[1] # ms -> s
    w = fitparam[2]
    SF = fitparam[3]
    Rcirc = fitparam[4]
    Tcirc = fitparam[5]
    alpha = 2 * np.pi / Tcirc * tau
    rho = Rcirc * np.sqrt(2-2*np.cos(alpha)) # = 2 * Rcirc * abs(sin(alpha/2))
    dc = fitparam[6]
    
    # get diffusion coefficient
    D = w**2 / 4 / tauD0

    yModel = fcs_dualfocus(tau, N, D, w, SF, rho, 0, dc, 0, 0)

    res = yexp - yModel
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_finitelength(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    Fit function finite length, 3D free diffusion
    Based on Kohler et al., Biophys. J., 2023

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, Tau (ms), Shape parameter, T (s), Tsampling (ms)]
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        same order as fitparamStart
    fit_info : 1D np.array
        np.array boolean vector with always 6 elements
        1 for a fitted parameter, 0 for a fixed parameter.
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array or 1, optional
        Vector with weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    fitparam = np.float64(np.zeros(6))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    
    #
    N = fitparam[0]
    tauD = 1e-3 * fitparam[1] # ms -> s
    SF = fitparam[2]
    T = fitparam[3] # segment length
    Tsampling = 1e-3 * fitparam[4] # sampling time (ms -> s)
    brightness = fitparam[5]
    
    FCStheo = fcs_analytical(tau, N, tauD, SF, 0, 0, 0, 1)
    gamma = compute_gamma(tau, N, tauD, T, Tsampling, gamma_factor=1, SP=SF, brightness=brightness)
    yModel = FCStheo + gamma
    
    res = yexp - yModel
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_free_diffusion_2d(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    fcs fit function two components 2D

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters:
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, offset, A, B]
        E.g. if only N and tauD1 are fitted, this becomes a two
        element vector [1, 1e-3].
        parameters  N       number of particles in focal volume [dim.less]
                    tauD1   diffusion time species 1 [ms]
                    tauD2   diffusion time species 2 [ms]
                    F       fraction of species 1 [dim.less]
                    alpha   relative molecular brightness [dim.less]
                    T       fraction in triplet [dim.less]
                    tautrip residence time in triplet state [µs]
                    offset  [dim.less]
                    A, B    Afterpulsing parameters
    fixedparam : 1D np.array
        List with values for the fixed parameters:
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, offset, A, B]
        same principle as fitparamStart.
    fit_info : 1D np.array
        np.array boolean vector with always 11 elements
        1 for a fitted parameter, 0 for a fixed parameter
        E.g. to fit N and tau D this becomes [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        order: [N, tauD1, tauD2, F, alpha, T, tautrip, offset, A, B].
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(10))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    # convert to SI
    N = fitparam[0]
    tauD1 = 1e-3 * fitparam[1] # ms to s
    tauD2 = 1e-3 * fitparam[2] # ms to s
    F = fitparam[3]
    alpha = fitparam[4]
    T = fitparam[5]
    tautrip = 1e-6 * fitparam[6] # s
    offset = fitparam[7]
    A = fitparam[8]
    B = fitparam[9]
    
    # calculate theoretical autocorrelation function
    FCStheo = fcs_2c_2d_analytical(tau, N, tauD1, tauD2, F, alpha, T, tautrip, offset, A, B)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_nanosecond_fcs(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    Fit with nanosecond fcs fit model
    See Galvanetto et al., Nature, 2023 (original formula from elsewhere)

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters
    fixedparam : 1D np.array
        List with values for the fixed parameters
    fit_info : 1D np.array
        np.array boolean vector with always 11 elements
        1 for a fitted parameter, 0 for a fixed parameter
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(11))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    # convert to SI
    A        = fitparam[0] # amplitude
    c_ab     = fitparam[1] # antibunching amplitude
    tau_ab   = fitparam[2] # antibunching time
    c_conf   = fitparam[3] # conformational dynamics
    tau_conf = fitparam[4]
    c_rot    = fitparam[5] # rotational dynamics
    tau_rot  = fitparam[6]
    c_trip   = fitparam[7] # triplet
    tau_trip = fitparam[8]
    tauD     = fitparam[9] # diffusion
    SP       = fitparam[10] # shape parameter
    
    
    # calculate theoretical autocorrelation function
    FCStheo = nanosecond_fcs_analytical(tau, A, c_ab, tau_ab, c_conf, tau_conf, c_rot, tau_rot, c_trip, tau_trip, tauD, SP)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res


def fitfun_uncoupled_reaction_diffusion(fitparamStart, fixedparam, fit_info, tau, yexp, weights=1):
    """
    Fit with uncoupled reaction and diffusion model
    Assumes that tauD << 1/k_on
    See Mazza et al., ch 12, Monitoring Dynamic Binding of Chromatin Proteins
    In Vivo by Fluorescence Correlation Spectroscopy
    and Temporal Image Correlation Spectroscopy

    Parameters
    ----------
    fitparamStart : 1D np.array
        List with starting values for the fit parameters
    fixedparam : 1D np.array
        List with values for the fixed parameters
    fit_info : 1D np.array
        np.array boolean vector with always 5 elements
        1 for a fitted parameter, 0 for a fixed parameter
    tau : 1D np.array
        Vector with tau values.
    yexp : 1D np.array
        Vector with experimental autocorrelation.
    weights : 1D np.array, optional
        Weights. The default is 1.

    Returns
    -------
    res : 1D np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(5))
    fitparam[fit_info==1] = fitparamStart
    fitparam[fit_info==0] = fixedparam
    
    # convert to SI
    A        = fitparam[0] # amplitude
    tauD     = fitparam[1] # diffusion (s)
    SP       = fitparam[2] # shape parameter
    f_eq     = fitparam[3] # fraction of free molecules at equilibrium
    k_off    = fitparam[4] # dissociation rate (/s)
    
    
    # calculate theoretical autocorrelation function
    FCStheo = uncoupled_reaction_diffusion(tau, A, tauD, SP, f_eq, k_off)
    
    # calculate residuals
    res = yexp - FCStheo
    
    # calculate weighted residuals
    res *= weights
    
    return res


def compute_gamma(tau, Nav, tau_D, T=10, T_s=10e-6, gamma_factor=0.51, SP=3, brightness=100):
    """
    Function needed for the finite length fit function   

    """
    # code by Lisa Cuneo, adapted to 3D by Eli (formula from Kohler et al, Eq. 26)
    
    k_medio = Nav * brightness * T_s
    
    # for 2D
    # B = lambda t, tau : 2 * (tau)**2 * ( (1+t) * np.log(1+t) - t)
    
    r = SP
    s = np.sqrt(r**2-1)
    
    # B2 (Eq. 26 Muller Biophys J. (86) 2004)
    B = lambda x, tD : 4 * r * tD**2 / s * (r*s - s*np.sqrt(r**2+x) - (1+x) * np.log((r-s)*(s+np.sqrt(r**2+x))/np.sqrt(1+x)))
    
    # Eq. S23
    kappa = lambda t : gamma_factor * brightness**2 * Nav * B(t / tau_D, tau_D)
    
    # Eq. S21
    Gamma_C = - T_s**2/(T-tau)**2 * ( kappa(T) + kappa(abs(T-2*tau)) - 2*kappa(tau) ) / ( 2*(k_medio**2) )
    
    #` Eq. 8 main text
    Gamma_S = - ( T_s * (T - 2 * tau) ) / ( (T - tau)**2 * k_medio )
    Gamma_S = np.where(T - 2*tau > 0, Gamma_S, 0 )
    
    Gamma = Gamma_C + Gamma_S
    
    return Gamma


def plot_fit(tau, yexp, param, fit_info, fitResult, plotInfo="", savefig=0, plotTau=True):  
    # final parameters
    param[fit_info==1] = fitResult.x
    # param now contains the final values for [N, tauD, SF, offset]
    # store in object
    data = PyCorrfit_data()
    if len(param) < 7:
        data.tauD = 1e3 * param[1] # in ms
    else:
        data.tauD = [param[1], param[2]] # in ms, 2 components
    
    G = yexp
    Gres = fitResult.fun
    Gfit = G - Gres
    
    Gdata = np.zeros([len(tau), 4])
    Gdata[:, 0] = tau
    Gdata[:, 1] = G
    Gdata[:, 2] = Gfit
    Gdata[:, 3] = Gres
    
    data.data = Gdata
    
    # plot
    plot_pycorrfit(data, plotInfo, savefig, plotTau)


def g2global_fit_struct(G, listOfFields=['spatialCorr'], start=0, N=5):
    """
    Create all variables needed for fcs fit based on correlation object G

    Parameters
    ----------
    G : object
        Correlation object.
    listOfFields : list, optional
        Type of analysis, e.g. "spatialCorr", "crossCenterAv",...
        The default is ['spatialCorr'].
    start : scalar, optional
        Start index for G and tau. The default is 0.
    N : scalar, optional
        Number of correlations to use
        For "spatialCorr", N=3 means the central 3x3 square of G
        For "crossCenterAv", N=3 means rho=1, rho=sqrt(2), and
            rho=2 are used. The default is 5.

    Returns
    -------
    Gout : 1D np.array
        Vector with all G's appended to one another.
    tau : 1D np.array
        Vector with all tau's appended to one another.
    split : 1D np.array
        Vector with the boundary indices for Gout and tau.
    rhox : 1D np.array
        Vector with rhox values for all Gout values.
    rhoy : 1D np.array
        Vector with rhoy values for all Gout values.
    Gcrop : 2D np.array
        2D matrix with all G's in different columns.

    """
    
    if listOfFields == '':
        listOfFields = list(G.__dict__.keys())
    pxdwelltime = G.dwellTime
    tau = []
    Gout = []
    rhox = []
    rhoy = []
    if 'spatialCorr' in listOfFields:
        Gcrop = G.spatialCorr[:, :, start:]
        Gshape = np.shape(Gcrop)
        GNy = Gshape[0]
        GNx = Gshape[1]
        GNt = Gshape[2]
        Gcolumns = np.zeros((GNt, N**2))
        Centerx = int(np.floor(GNx/2))
        Centery = int(np.floor(GNy/2))
        GxStart = int(-np.floor(N/2))
        for i in range(N):
            for j in range(N):
                Gcolumns[:, i*N+j] = Gcrop[GxStart+i+Centery, GxStart+j+Centerx, :]
                rhox = np.append(rhox, GxStart+j)
                rhoy = np.append(rhoy, GxStart+i)
                tau = np.append(tau, np.linspace(start*pxdwelltime, pxdwelltime*(GNt-1), GNt-start))
                Gout = np.append(Gout, Gcrop[i, j, :])
        Gcrop = Gcolumns
        split = np.linspace(0, GNt*N*N, N*N+1)
    elif 'crossCenterAv' in listOfFields:
        # average all pair-correlations between the center and the other pixels
        # that are equally far from the center
        Gcrop = G.crossCenterAv[start:,:]
        Gshape = np.shape(Gcrop)
        GNt = Gshape[0]
        for i in range(6):
            Gout = np.append(Gout, Gcrop[:, i])
            tau = np.append(tau, np.linspace(start*pxdwelltime, pxdwelltime*(GNt-1), GNt-start))
        rhox = np.array([0, 1, np.sqrt(2), 2, np.sqrt(5), np.sqrt(8)])
        rhoy = np.array([0, 0, 0, 0, 0, 0])
        split = np.linspace(0, GNt*6, 7)
    elif 'crossCenterAvN' in listOfFields:
        # similar to crossCenterAv, but without the central autocorrelation
        # only the N G's closest to the center are calculated
        Gcrop = G.crossCenterAv[start:,1:N+1]
        Gshape = np.shape(Gcrop)
        GNt = Gshape[0]
        for i in range(N):
            Gout = np.append(Gout, Gcrop[:, i])
            tau = np.append(tau, np.linspace(start*pxdwelltime, pxdwelltime*(GNt-1), GNt-start))
        rhox = np.array([1, np.sqrt(2), 2, np.sqrt(5), np.sqrt(8)])
        rhox = rhox[0:N]
        rhoy = np.array([0, 0, 0, 0, 0])
        rhoy = rhoy[0:N]
        split = np.linspace(0, GNt*N, N+1)
    elif 'twofocus5' in listOfFields:
        # average all pair-correlations between the center and the other pixels
        # that are equally far from the center
        Nt = len(G.twofocusTB_average[start:,0])
        Gcrop = np.zeros((Nt, 6))
        # cross-correlation top and bottom
        Gcrop[:,0] = (G.twofocusTB_average[start:,1] + G.twofocusBT_average[start:,1]) / 2
        # cross-correlation left and right
        Gcrop[:,1] = (G.twofocusLR_average[start:,1] + G.twofocusRL_average[start:,1]) / 2
        # cross-correlation left and top
        Gcrop[:,2] = (G.twofocusLT_average[start:,1] + G.twofocusTL_average[start:,1]) / 2
        # cross-correlation bottom and right
        Gcrop[:,3] = (G.twofocusBR_average[start:,1] + G.twofocusRB_average[start:,1]) / 2
        # cross-correlation left and bottom
        Gcrop[:,4] = (G.twofocusLB_average[start:,1] + G.twofocusBL_average[start:,1]) / 2
        # cross-correlation top and right
        Gcrop[:,5] = (G.twofocusTR_average[start:,1] + G.twofocusRT_average[start:,1]) / 2
        for i in range(6):
            Gout = np.append(Gout, Gcrop[:, i])
            tau = np.append(tau, np.linspace(start*pxdwelltime, pxdwelltime*(Nt-1), Nt-start))
        rhox = np.array([2, 2, np.sqrt(2), np.sqrt(2), np.sqrt(2), np.sqrt(2)])
        rhoy = np.array([0, 0, 0, 0, 0, 0])
        split = np.linspace(0, Nt*6, 7)
    elif 'twofocus5Av' in listOfFields:
        # average all pair-correlations between the center and the other pixels
        # that are equally far from the center
        Nt = len(G.twofocusTB_average[start:,0])
        Gcrop = np.zeros((Nt, 2))
        # cross-correlation long distance
        Gcrop[:,0] = G.twofocusTB_average[start:,1] + G.twofocusBT_average[start:,1]
        Gcrop[:,0] += G.twofocusLR_average[start:,1] + G.twofocusRL_average[start:,1]
        Gcrop[:,0] /= 4
        # cross-correlation short distance
        Gcrop[:,1] = G.twofocusLT_average[start:,1] + G.twofocusTL_average[start:,1]
        Gcrop[:,1] += G.twofocusBR_average[start:,1] + G.twofocusRB_average[start:,1]
        Gcrop[:,1] += G.twofocusLB_average[start:,1] + G.twofocusBL_average[start:,1]
        Gcrop[:,1] += G.twofocusTR_average[start:,1] + G.twofocusRT_average[start:,1]
        Gcrop[:,1] /= 8
        for i in range(2):
            Gout = np.append(Gout, Gcrop[:, i])
            tau = np.append(tau, np.linspace(start*pxdwelltime, pxdwelltime*(Nt-1), Nt-start))
        rhox = np.array([2, np.sqrt(2)])
        rhoy = np.array([0, 0])
        split = np.linspace(0, Nt*2, 3)
    return Gout, tau, split, rhox, rhoy, Gcrop


def chi_2_fcs_fit(tau, Gexp, fit_info, fitResult):
    # calculate the chi2 value for the fcs fit
    Gres = fitResult.fun
    Gfit = Gexp - Gres
    dof = len(tau) - np.sum(fit_info) - 1 # degrees of freedom
    chi2 = np.sum((Gexp-Gfit)**2 / np.abs(Gfit)) / dof
    return chi2
    


def constants(c):
    """
    Physical constants

    Parameters
    ----------
    c : string
        name of the constant.

    Returns
    -------
    float
        constant value.

    """
    
    c = c.lower()
    if c == 'boltzmann':
        return 1.38064852e-23  # m^2 kg / (K s^2)
    elif c == 'planck':
        return 6.62607004e-35  # m^2 kg / s
    elif c == 'avogadro':
        return 6.02214076e23 # no units
    else:
        return 0



# -*- coding: utf-8 -*-
"""
Change current working directory to the folder given as input
"""

import os
import platform

def goto(directory):
    if directory == 'data':
        path = 'SPAD-fcs\\Data\\'
    elif directory == 'fcs':
        path = 'Python\\fcs\\'
    elif directory == 'tools':
        path = 'Python\\tools\\'
    else:
        path = directory
    
    pf = platform.node()
    if pf == 'IITLW1768':
        # IIT laptop
        basePath = 'C:\\Users\\eslenders\\OneDrive\\OneDrive - Fondazione Istituto Italiano Tecnologia\\'
    elif pf == 'IITMMSDW002':
        # IIT desktop pc
        basePath = 'C:\\Users\\eslenders\\OneDrive - Fondazione Istituto Italiano Tecnologia\\'
    else:
        # lab PC
        basePath = 'C:\\Users\\SPAD-fcs\\OneDrive - Fondazione Istituto Italiano Tecnologia\\'
        
    os.chdir(basePath + path)
